# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import typing

from pathlib import Path

import gi

try:
    gi.require_version("Gst", "1.0")
    from gi.repository import Gst
except Exception:
    if typing.TYPE_CHECKING:
        from gi.repository import Gst

Gst.init(None)


def extract_video_thumbnail_and_properties(
    input_: Path, output: Path | None, preview_size: int
) -> tuple[bytes, dict[str, typing.Any]]:
    pipeline = Gst.Pipeline.new()

    uridecodebin = Gst.ElementFactory.make("uridecodebin3")
    videoconvert = Gst.ElementFactory.make("videoconvert")
    videoscale = Gst.ElementFactory.make("videoscale")
    capsfilter = Gst.ElementFactory.make("capsfilter")
    pngenc = Gst.ElementFactory.make("pngenc")
    appsink = Gst.ElementFactory.make("appsink")

    pipeline_elements = [
        uridecodebin,
        videoconvert,
        videoscale,
        capsfilter,
        pngenc,
        appsink,
    ]
    if any(element is None for element in pipeline_elements):
        raise Exception(f"\n{__name__}: Some pipeline elements were None")

    assert uridecodebin is not None
    assert videoconvert is not None
    assert videoscale is not None
    assert capsfilter is not None
    assert pngenc is not None
    assert appsink is not None

    appsink.set_property("emit-signals", False)
    appsink.set_property("sync", False)
    appsink.set_property("max-buffers", 1)
    appsink.set_property("drop", True)
    uridecodebin.set_property("uri", input_.as_uri())

    pipeline.add(uridecodebin)
    pipeline.add(videoconvert)
    pipeline.add(videoscale)
    pipeline.add(capsfilter)
    pipeline.add(pngenc)
    pipeline.add(appsink)

    metadata: dict[str, typing.Any] = {"width": 0, "height": 0}

    def probe_original_size(
        pad: Gst.Pad, _info: Gst.PadProbeInfo
    ) -> Gst.PadProbeReturn:
        caps = pad.get_current_caps()
        if caps is None:
            return Gst.PadProbeReturn.OK

        structure = caps.get_structure(0)
        if structure.has_field("width") and structure.has_field("height"):
            metadata["width"] = structure.get_int("width")[1]
            metadata["height"] = structure.get_int("height")[1]
            return Gst.PadProbeReturn.REMOVE

        return Gst.PadProbeReturn.OK

    sink_pad = videoscale.get_static_pad("sink")
    assert sink_pad is not None
    sink_pad.add_probe(Gst.PadProbeType.EVENT_DOWNSTREAM, probe_original_size)

    def on_pad_added(_bin: Gst.Bin, pad: Gst.Pad) -> None:
        assert pad is not None
        sink_pad = videoconvert.get_static_pad("sink")
        assert sink_pad is not None
        if not sink_pad.is_linked():
            pad.link(sink_pad)

    handler_id = uridecodebin.connect("pad-added", on_pad_added)

    videoconvert.link(videoscale)
    videoscale.link(capsfilter)
    capsfilter.link(pngenc)
    pngenc.link(appsink)

    # https://gitlab.freedesktop.org/gstreamer/gstreamer/-/blob/1.26/subprojects/gst-plugins-base/gst/videoconvertscale/gstvideoconvertscale.h#L63
    lanczos_filter = 3
    videoscale.set_property("method", lanczos_filter)
    caps = Gst.Caps.from_string(
        f"video/x-raw,width={preview_size},pixel-aspect-ratio=1/1"
    )
    capsfilter.set_property("caps", caps)

    pipeline.set_state(Gst.State.PAUSED)

    def cleanup() -> None:
        pipeline.set_state(Gst.State.NULL)
        uridecodebin.disconnect(handler_id)
        for elem in pipeline_elements:
            if elem is not None:
                pipeline.remove(elem)
                elem.set_state(Gst.State.NULL)
        pipeline.run_dispose()

    state_change, _, _ = pipeline.get_state(Gst.CLOCK_TIME_NONE)
    if state_change != Gst.StateChangeReturn.SUCCESS:
        cleanup()
        raise Exception(f"\n{__name__}: State change was not successful")

    success, duration_ns = pipeline.query_duration(Gst.Format.TIME)
    if not success:
        duration_ns = 0

    # Take timestamp after 2 seconds or earlier, if duration is shorter
    duration_ms = duration_ns / 1e6
    metadata["duration"] = duration_ms
    timestamp_ms = min(2000, int(0.5 * duration_ms))

    pipeline.seek_simple(
        Gst.Format.TIME,
        Gst.SeekFlags.FLUSH | Gst.SeekFlags.ACCURATE,
        int(timestamp_ms) * Gst.MSECOND,
    )

    pipeline.set_state(Gst.State.PLAYING)
    sample = appsink.emit("try-pull-sample", 2 * Gst.SECOND)

    if sample is None:
        cleanup()
        raise Exception(f"\n{__name__}: Failed to retrieve sample")

    buffer = sample.get_buffer()
    success, mapinfo = buffer.map(Gst.MapFlags.READ)

    if not success:
        cleanup()
        raise Exception(f"\n{__name__}: Failed to map buffer")

    bytes_ = bytes(mapinfo.data)
    buffer.unmap(mapinfo)

    cleanup()

    if output is not None:
        output.write_bytes(bytes_)
    return bytes_, metadata



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from gi.repository import Gtk

from gajim.common.const import AvatarSize
from gajim.common.const import EncryptionInfoMsg
from gajim.common.events import EncryptionInfo
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.util.datetime import utc_now

from gajim.gtk.conversation.rows.base import BaseRow
from gajim.gtk.conversation.rows.widgets import DateTimeLabel
from gajim.gtk.conversation.rows.widgets import SimpleLabel
from gajim.gtk.util.window import open_window


class EncryptionInfoRow(BaseRow):
    def __init__(self, event: EncryptionInfo) -> None:
        BaseRow.__init__(self, event.account)

        self.type = "encryption_info"
        timestamp = utc_now()
        self.timestamp = timestamp.astimezone()
        self._event = event

        avatar_placeholder = Gtk.Box()
        avatar_placeholder.set_size_request(AvatarSize.ROSTER, -1)

        icon = Gtk.Image.new_from_icon_name("lucide-lock-symbolic")
        icon.set_pixel_size(AvatarSize.ROSTER)
        icon.add_css_class("dimmed")
        avatar_placeholder.append(icon)
        self.grid.attach(avatar_placeholder, 0, 0, 1, 1)

        timestamp_widget = DateTimeLabel(self.timestamp)
        timestamp_widget.set_valign(Gtk.Align.START)
        timestamp_widget.set_margin_start(0)
        self.grid.attach(timestamp_widget, 1, 0, 1, 1)

        self._label = SimpleLabel()
        self._label.set_text(event.message)
        self.grid.attach(self._label, 1, 1, 1, 1)

        if event.type in (
            EncryptionInfoMsg.NO_FINGERPRINTS,
            EncryptionInfoMsg.UNDECIDED_FINGERPRINTS,
        ):
            button = Gtk.Button(label=_("Manage Trust"))
            button.set_halign(Gtk.Align.START)
            self._connect(button, "clicked", self._on_manage_trust_clicked)
            self.grid.attach(button, 1, 2, 1, 1)

    def do_unroot(self) -> None:
        BaseRow.do_unroot(self)

    def _on_manage_trust_clicked(self, _button: Gtk.Button) -> None:
        contact = self._client.get_module("Contacts").get_contact(self._event.jid)
        if contact.is_groupchat:
            open_window("GroupchatDetails", contact=contact, page="encryption-omemo")
            return

        if isinstance(contact, BareContact) and contact.is_self:
            window = open_window("Preferences")
            window.show_page(f"{contact.account}-encryption-omemo")
            return

        open_window(
            "ContactInfo",
            account=contact.account,
            contact=contact,
            page="encryption-omemo",
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Literal

import logging
from collections.abc import Iterator
from datetime import datetime

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp import JID

from gajim.common import app
from gajim.common import events
from gajim.common import ged
from gajim.common.const import Direction
from gajim.common.const import RowHeaderType
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.message_util import get_nickname_from_message
from gajim.common.setting_values import OpenChatsSettingT
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageType
from gajim.common.util.muc import get_group_chat_nick
from gajim.common.util.user_strings import get_moderation_text
from gajim.common.util.user_strings import get_retraction_text

from gajim.gtk.chat_filter import ChatFilters
from gajim.gtk.chat_list_row import ChatListRow
from gajim.gtk.preview.preview import PreviewWidget
from gajim.gtk.start_chat import ChatTypeFilter
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_listbox_row_count
from gajim.gtk.util.misc import iterate_listbox_children

log = logging.getLogger("gajim.gtk.chatlist")


class ChatList(Gtk.ListBox, EventHelper, SignalManager):
    __gsignals__ = {
        "chat-order-changed": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, workspace_id: str) -> None:
        Gtk.ListBox.__init__(self)
        EventHelper.__init__(self)
        SignalManager.__init__(self)

        self._workspace_id = workspace_id

        self._chats: dict[tuple[str, JID], ChatListRow] = {}
        self._current_filter: ChatFilters = ChatFilters()
        self._current_filter_text: str = ""

        self.add_css_class("chatlist")
        self.set_filter_func(self._filter_func)
        self.set_header_func(self._header_func)
        self.set_sort_func(self._sort_func)
        self._set_placeholder()

        self._force_sort = False
        self._rows_need_sort = False
        self._context_menu_visible = False
        self._mouseover = False
        self._scheduled_sort_id = None

        hover_controller = Gtk.EventControllerMotion()
        self._connect(hover_controller, "enter", self._on_cursor_enter)
        self._connect(hover_controller, "leave", self._on_cursor_leave)
        self.add_controller(hover_controller)

        chat_drop_target = Gtk.DropTarget(
            formats=Gdk.ContentFormats.new_for_gtype(ChatListRow),
            actions=Gdk.DragAction.MOVE,
        )
        self._connect(chat_drop_target, "drop", self._on_chat_drop)
        self.add_controller(chat_drop_target)

        file_drop_target = Gtk.DropTarget(
            formats=Gdk.ContentFormats.new_for_gtype(PreviewWidget),
            actions=Gdk.DragAction.COPY,
        )
        self._connect(file_drop_target, "drop", self._on_file_drop)
        self.add_controller(file_drop_target)

        self._chat_order: list[ChatListRow] = []

        self.register_events(
            [
                ("account-enabled", ged.GUI2, self._on_account_changed),
                ("account-disabled", ged.GUI2, self._on_account_changed),
                ("bookmarks-received", ged.GUI1, self._on_bookmarks_received),
            ]
        )

        app.pulse_manager.add_callback(self._update_row_state)

    def do_unroot(self) -> None:
        Gtk.ListBox.do_unroot(self)
        self._disconnect_all()
        self.unregister_events()
        self._abort_scheduled_sort("unroot")
        self.set_filter_func(None)
        self.set_header_func(None)
        self.set_sort_func(None)
        app.pulse_manager.remove_callback(self._update_row_state)
        app.check_finalize(self)

    @property
    def workspace_id(self) -> str:
        return self._workspace_id

    def get_row_count(self) -> int:
        return get_listbox_row_count(self)

    def _iterate_rows(self) -> Iterator[ChatListRow]:
        for row in iterate_listbox_children(self):
            yield cast(ChatListRow, row)

    def get_unread_count(self) -> int:
        return sum(chats.unread_count for chats in self._chats.values())

    def get_chat_unread_count(
        self, account: str, jid: JID, include_silent: bool = False
    ) -> int | None:
        chat = self._chats.get((account, jid))
        if chat is not None:
            if include_silent:
                return chat.get_real_unread_count()
            return chat.unread_count
        return None

    def set_chat_unread_count(self, account: str, jid: JID, count: int) -> None:
        chat = self._chats.get((account, jid))
        if chat is not None:
            chat.unread_count = count

    def set_filter(self, chat_filter: ChatFilters) -> None:
        self._current_filter = chat_filter
        self.invalidate_filter()

    def set_filter_text(self, text: str) -> None:
        self._current_filter_text = text
        self.invalidate_filter()

    def get_chat_type(
        self, account: str, jid: JID
    ) -> Literal["chat", "groupchat", "pm"] | None:
        row = self._chats.get((account, jid))
        if row is not None:
            return row.type  # type: ignore
        return None

    def get_selected_chat(self) -> ChatListRow | None:
        row = cast(ChatListRow | None, self.get_selected_row())
        if row is None:
            return None
        return row

    def get_open_chats(self) -> OpenChatsSettingT:
        open_chats: OpenChatsSettingT = []
        for key, row in self._chats.items():
            account, jid = key
            open_chats.append(
                {
                    "account": account,
                    "jid": jid,
                    "type": row.type,  # type: ignore
                    "pinned": row.is_pinned,
                    "position": row.position,
                }
            )
        return open_chats

    def mark_as_read(self, account: str, jid: JID) -> None:
        chat = self._chats.get((account, jid))
        if chat is not None and chat.get_real_unread_count() > 0:
            chat.reset_unread()
            app.ged.raise_event(events.ChatRead(account=account, jid=jid))

    def toggle_chat_pinned(self, account: str, jid: JID) -> None:
        row = self._chats[(account, jid)]

        if row.is_pinned:
            self._chat_order.remove(row)
            row.position = -1
        else:
            self._chat_order.append(row)
            row.position = self._chat_order.index(row)

        row.toggle_pinned()
        self.invalidate_sort(force=True)

    def add_chat(
        self,
        account: str,
        jid: JID,
        type_: Literal["chat", "groupchat", "pm"],
        pinned: bool,
        position: int,
    ) -> None:
        key = (account, jid)
        if self._chats.get(key) is not None:
            # Chat is already in the List
            return

        row = ChatListRow(self._workspace_id, account, jid, type_, pinned, position)

        self._chats[key] = row
        if pinned:
            self._chat_order.insert(position, row)

        row.connect("unread-changed", self._on_row_unread_changed)
        row.connect("context-menu-state-changed", self._on_context_menu_state_changed)

        self.append(row)

    def select_chat(self, account: str, jid: JID) -> None:
        row = self._chats[(account, jid)]
        self.select_row(row)

    def select_next_chat(
        self, direction: Direction, unread_first: bool = False
    ) -> None:
        # Selects the next chat, but prioritizes chats with unread messages.
        row = self.get_selected_chat()
        if row is None:
            row = self.get_row_at_index(0)
            if row is None:
                return
            assert isinstance(row, ChatListRow)
            self.select_chat(row.account, row.jid)
            return

        unread_found = False
        if unread_first:
            index = row.get_index()
            current = index

            # Loop until finding a chat with unread count or completing a cycle
            while True:
                if direction == Direction.NEXT:
                    index += 1
                    if index >= self.get_row_count():
                        index = 0
                else:
                    index -= 1
                    if index < 0:
                        index = self.get_row_count() - 1

                row = self.get_row_at_index(index)
                if row is None:
                    return
                assert isinstance(row, ChatListRow)
                if row.unread_count > 0:
                    unread_found = True
                    break
                if index == current:
                    break

        if unread_found:
            self.select_chat(row.account, row.jid)
            return

        index = row.get_index()
        if direction == Direction.NEXT:
            next_row = self.get_row_at_index(index + 1)
        else:
            next_row = self.get_row_at_index(index - 1)
        if next_row is None:
            if direction == Direction.NEXT:
                next_row = self.get_row_at_index(0)
            else:
                last = self.get_row_count() - 1
                next_row = self.get_row_at_index(last)
            assert isinstance(next_row, ChatListRow)
            self.select_chat(next_row.account, next_row.jid)
            return

        assert isinstance(next_row, ChatListRow)
        self.select_chat(next_row.account, next_row.jid)

    def select_chat_number(self, number: int) -> None:
        row = self.get_row_at_index(number)
        if row is not None:
            assert isinstance(row, ChatListRow)
            self.select_chat(row.account, row.jid)

    def get_chat_list_rows(self) -> list[ChatListRow]:
        return list(self._iterate_rows())

    def remove_chat(self, account: str, jid: JID, emit_unread: bool = True) -> None:
        row = self._chats.pop((account, jid))

        if row.is_pinned:
            self._chat_order.remove(row)
        self.remove(row)

        if emit_unread:
            self._emit_unread_changed()

    def remove_chats_for_account(self, account: str) -> None:
        for row_account, jid in list(self._chats.keys()):
            if row_account != account:
                continue
            self.remove_chat(account, jid)
        self._emit_unread_changed()

    def clear_chat_list_row(self, account: str, jid: JID) -> None:
        chat = self._chats.get((account, jid))
        if chat is not None:
            chat.clear()

    def contains_chat(self, account: str, jid: JID) -> bool:
        return self._chats.get((account, jid)) is not None

    def process_event(self, event: events.ChatListEventT) -> None:
        if isinstance(event, events.MessageReceived):
            self._on_message_received(event)
        elif isinstance(event, events.MessageDeleted):
            self._on_message_deleted(event)
        elif isinstance(event, events.MessageCorrected):
            self._on_message_corrected(event)
        elif isinstance(event, events.MessageModerated):
            self._on_message_moderated(event)
        elif isinstance(event, events.MessageRetracted):
            self._on_message_retracted(event)
        elif isinstance(event, events.PresenceReceived):
            self._on_presence_received(event)
        elif isinstance(event, events.MessageSent):
            self._on_message_sent(event)
        elif isinstance(event, events.JingleRequestReceived):
            self._on_jingle_request_received(event)
        elif isinstance(event, events.FileRequestReceivedEvent):  # pyright: ignore [reportUnnecessaryIsInstance] # noqa: E501
            self._on_file_request_received(event)
        else:
            log.warning("Unhandled Event: %s", event.name)

    def _set_placeholder(self) -> None:
        button = Gtk.Button(
            label=_("Start Chat"),
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
            action_name="app.start-chat",
            action_target=GLib.Variant("as", ["", ""]),
        )
        button.add_css_class("suggested-action")
        self.set_placeholder(button)

    def _emit_unread_changed(self) -> None:
        count = self.get_unread_count()
        chat_list_stack = self.get_parent()
        assert chat_list_stack is not None
        chat_list_stack.emit("unread-count-changed", self._workspace_id, count)

    def _get_row_before(self, row: ChatListRow) -> ChatListRow | None:
        row_before = self.get_row_at_index(row.get_index() - 1)
        if row_before is None:
            return
        return cast(ChatListRow, row_before)

    def _get_row_after(self, row: ChatListRow) -> ChatListRow | None:
        row_after = self.get_row_at_index(row.get_index() + 1)
        if row_after is None:
            return
        return cast(ChatListRow, row_after)

    def _get_last_row(self) -> ChatListRow:
        index = self.get_row_count() - 1
        last_row = self.get_row_at_index(index)
        assert last_row is not None
        return cast(ChatListRow, last_row)

    def _on_row_unread_changed(self, row: ChatListRow) -> None:
        self._emit_unread_changed()

    def _on_context_menu_state_changed(
        self, row: ChatListRow, menu_is_visible: bool
    ) -> None:
        self._context_menu_visible = menu_is_visible
        self._schedule_sort()

    def _on_file_drop(
        self, _drop_target: Gtk.DropTarget, value: GObject.Value, _x: float, y: float
    ) -> bool:
        target_row = self.get_row_at_y(int(y))
        if not value or not isinstance(target_row, ChatListRow):
            # Reject drop
            return False

        if isinstance(value, PreviewWidget):
            # Calling highlight_dnd_targets False is necessary here because there is no
            # drag-end signal in this case
            app.window.highlight_dnd_targets(value, False)
            app.window.select_chat(target_row.account, target_row.jid)

            uri: list[str] = [value.get_orig_path().as_uri()]
            app.window.activate_action("win.send-file", GLib.Variant("as", uri))
            return True

        # Reject drop
        log.debug("Unknown item type dropped")
        return False

    def _on_chat_drop(
        self, _drop_target: Gtk.DropTarget, value: GObject.Value, _x: float, y: float
    ) -> bool:
        target_row = self.get_row_at_y(int(y))
        if not value or not isinstance(target_row, ChatListRow):
            # Reject drop
            return False

        if isinstance(value, ChatListRow):
            if not value.is_pinned or not target_row.is_pinned:
                log.debug("Only pinned rows can be reordered")
                return False

            self._change_pinned_order(value, target_row)
            return True

        # Reject drop
        log.debug("Unknown item type dropped")
        return False

    def _change_pinned_order(
        self, drag_row: ChatListRow, target_row: ChatListRow
    ) -> None:
        self._chat_order.remove(drag_row)
        self._chat_order.insert(target_row.position, drag_row)

        for row in self._chat_order:
            row.position = self._chat_order.index(row)

        self.emit("chat-order-changed")
        self.invalidate_sort(force=True)

    def _update_row_state(self) -> bool:
        for row in self._chats.values():
            row.update_row_state()
        return True

    def _filter_func(self, row: ChatListRow) -> bool:
        account = self._current_filter.account
        if account is not None and account != row.account:
            return False

        is_groupchat = row.type == "groupchat"
        is_pm = row.type == "pm"
        if self._current_filter.type == ChatTypeFilter.CHAT and is_groupchat:
            return False

        if self._current_filter.type == ChatTypeFilter.GROUPCHAT and not is_groupchat:
            return False

        group = self._current_filter.group
        if group is not None:
            if is_groupchat or is_pm:
                return False

            assert isinstance(row.contact, BareContact)
            if group not in row.contact.groups:
                return False

        if not self._current_filter_text:
            return True

        text = self._current_filter_text.lower()
        return text in row.contact_name.lower()

    @staticmethod
    def _header_func(row: ChatListRow, before: ChatListRow | None) -> None:
        if before is None:
            if row.is_pinned:
                row.set_header_type(RowHeaderType.PINNED)
            else:
                row.set_header_type(None)
        else:
            if row.is_pinned:
                if before.is_pinned:
                    row.set_header_type(None)
                else:
                    row.set_header_type(RowHeaderType.PINNED)
            else:
                if before.is_pinned:
                    row.set_header_type(RowHeaderType.CONVERSATIONS)
                else:
                    row.set_header_type(None)

    def _sort_func(self, row1: ChatListRow, row2: ChatListRow) -> int:
        if not self._force_sort and self._is_sort_inhibited():
            self._rows_need_sort = True
            log.debug("Delay sorting because it is inhibited")
            return 0

        if row1.is_pinned and row2.is_pinned:
            return -1 if row1.position < row2.position else 1

        if row1.is_pinned != row2.is_pinned:
            return -1 if row1.is_pinned else 1

        if row1.has_draft != row2.has_draft:
            return -1 if row1.has_draft else 1

        if row1.timestamp != row2.timestamp:
            return -1 if row1.timestamp > row2.timestamp else 1

        return 0

    def invalidate_sort(self, *, force: bool = False) -> bool:  # pyright: ignore
        log.debug("Try sorting chatlist")
        if not force and self._is_sort_inhibited():
            log.debug("Abort sorting because it is inhibited")
            return False

        self._rows_need_sort = False
        self._force_sort = force
        Gtk.ListBox.invalidate_sort(self)
        self._force_sort = False
        log.debug("Sorting successful")
        return True

    def _is_sort_inhibited(self) -> bool:
        return self._mouseover or self._context_menu_visible

    def _schedule_sort(self) -> None:
        self._abort_scheduled_sort("schedule is renewed")
        log.debug("Schedule sort")
        self._scheduled_sort_id = GLib.timeout_add(100, self._execute_scheduled_sort)

    def _abort_scheduled_sort(self, reason: str) -> None:
        if self._scheduled_sort_id is not None:
            log.debug("Abort scheduled sort, reason: %s", reason)
            GLib.source_remove(self._scheduled_sort_id)
        self._scheduled_sort_id = None

    def _execute_scheduled_sort(self) -> bool:
        log.debug("Execute scheduled sort")
        if not self._rows_need_sort:
            log.debug("Abort scheduled sort, reason: no rows changed")
            self._scheduled_sort_id = None
            return GLib.SOURCE_REMOVE

        sort_executed = self.invalidate_sort()
        if sort_executed:
            self._scheduled_sort_id = None
            return GLib.SOURCE_REMOVE

        return GLib.SOURCE_CONTINUE

    def _on_cursor_enter(
        self,
        _controller: Gtk.EventControllerMotion,
        _x: int,
        _y: int,
    ) -> None:
        self._mouseover = True
        self._abort_scheduled_sort("mouse over list")

    def _on_cursor_leave(self, _controller: Gtk.EventControllerMotion) -> None:
        self._mouseover = False
        self._schedule_sort()

    @staticmethod
    def _add_unread(row: ChatListRow, event: events.MessageReceived) -> None:
        message = event.message
        if message.direction == ChatDirection.OUTGOING:
            # Last message was from us (1:1), reset counter
            row.reset_unread()
            return

        our_nick = get_group_chat_nick(event.account, event.jid)
        if message.resource == our_nick:
            # Last message was from us (MUC), reset counter
            row.reset_unread()
            return

        control = app.window.get_control()
        if app.window.is_active() and row.is_selected() and control.view_is_at_bottom():
            return

        assert message.text is not None
        row.add_unread(message.text)

    def _on_message_received(self, event: events.MessageReceived) -> None:
        row = self._chats.get((event.account, JID.from_string(event.jid)))
        if row is None:
            return

        message = event.message
        if message.text is None:
            return

        if message.type == MessageType.GROUPCHAT and message.occupant is not None:
            client = app.get_client(event.account)
            if client.get_module("MucBlocking").is_blocked(
                message.remote.jid, message.occupant.id
            ):
                return

        assert message.id is not None

        nick, text_nick = get_nickname_from_message(event.account, message)
        row.set_nick(nick)
        row.set_timestamp(message.timestamp)
        row.set_stanza_id(message.stanza_id)
        row.set_message_id(message.id)
        row.set_message_text(message.text, nickname=text_nick, oob=message.oob)

        self._add_unread(row, event)
        row.changed()

    def _on_message_deleted(self, event: events.MessageDeleted) -> None:
        # TODO
        pass

    def _on_message_corrected(self, event: events.MessageCorrected) -> None:
        row = self._chats.get((event.account, JID.from_string(event.jid)))
        if row is None:
            return

        if event.correction_id == row.message_id:
            text = event.message.text
            assert text is not None
            _, text_nick = get_nickname_from_message(event.account, event.message)
            row.set_message_text(text, nickname=text_nick)

    def _on_message_moderated(self, event: events.MessageModerated) -> None:
        row = self._chats.get((event.account, event.jid))
        if row is None:
            return

        if event.moderation.stanza_id == row.stanza_id:
            text = get_moderation_text(event.moderation.by, event.moderation.reason)
            row.set_message_text(text)

    def _on_message_retracted(self, event: events.MessageRetracted) -> None:
        row = self._chats.get((event.account, event.jid))
        if row is None:
            return

        if event.retraction.id in (row.message_id, row.stanza_id):
            row.set_message_text(get_retraction_text(event.retraction.timestamp))

    def _on_message_sent(self, event: events.MessageSent) -> None:
        row = self._chats.get((event.account, JID.from_string(event.jid)))
        if row is None:
            return

        message = event.message
        assert message.text is not None

        row.set_nick(_("Me"))
        row.set_timestamp(message.timestamp)
        row.set_message_text(
            message.text, nickname=app.nicks[event.account], oob=message.oob
        )
        row.changed()

    def _on_presence_received(self, event: events.PresenceReceived) -> None:
        row = self._chats.get((event.account, JID.from_string(event.jid)))
        if row is None:
            return
        row.update_avatar()

    def _on_jingle_request_received(self, event: events.JingleRequestReceived) -> None:
        content_types: list[str] = []
        for item in event.contents:
            content_types.append(item.media)
        if "audio" in content_types or "video" in content_types:
            # AV Call received
            row = self._chats.get((event.account, JID.from_string(event.jid)))
            if row is None:
                return
            row.set_timestamp(datetime.now().astimezone())
            row.set_nick("")
            row.set_message_text(_("Call"), icon_name="lucide-phone-symbolic")

    def _on_file_request_received(self, event: events.FileRequestReceivedEvent) -> None:
        row = self._chats.get((event.account, event.jid))
        if row is None:
            return
        row.set_timestamp(datetime.now().astimezone())
        row.set_nick("")
        row.set_message_text(_("File"), icon_name="lucide-paperclip-symbolic")

    def _on_account_changed(self, *args: Any) -> None:
        for row in self._iterate_rows():
            row.update_account_identifier()

    def _on_bookmarks_received(self, _event: events.BookmarksReceived) -> None:
        for row in self._iterate_rows():
            row.update_name()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import NamedTuple

import logging
import sys
from dataclasses import dataclass
from pathlib import Path
from urllib.parse import ParseResult
from urllib.parse import unquote
from urllib.parse import urlparse

from gi.repository import Gio
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common.const import NONREGISTERED_URI_SCHEMES
from gajim.common.iana.uri_schemes import URI_SCHEMES
from gajim.common.regex import IRI_RX

log = logging.getLogger("gajim.c.util.uri")


@dataclass
class Uri:
    scheme: str
    uri: str

    @classmethod
    def from_urlparts(cls, urlparts: ParseResult, uri: str) -> Uri:
        if urlparts.scheme in ("http", "https"):
            if not urlparts.netloc:
                raise ValueError(f"No host or empty host in an {urlparts.scheme} URI")
        return cls(scheme=urlparts.scheme, uri=uri)


@dataclass
class InvalidUri(Uri):
    error: str


@dataclass
class XmppIri(Uri):
    jid: JID
    action: str
    params: dict[str, str]

    @classmethod
    def from_urlparts(cls, urlparts: ParseResult, uri: str) -> XmppIri:
        urlparts = urlparse(uri)
        if urlparts.scheme != "xmpp":
            raise ValueError("Invalid scheme: %s" % urlparts.scheme)

        if urlparts.path.startswith("/"):
            raise ValueError("Authority component not supported: %s" % uri)

        if not urlparts.path:
            raise ValueError("No path component found: %s" % uri)

        jid = JID.from_iri(f"xmpp:{urlparts.path}")
        qtype, qparams = parse_xmpp_uri_query(urlparts.query)
        return cls(scheme="xmpp", uri=uri, jid=jid, action=qtype, params=qparams)

    @classmethod
    def from_string(cls, uri: str) -> XmppIri:
        urlparts = urlparse(uri)
        return XmppIri.from_urlparts(urlparts, uri)


@dataclass
class MailUri(Uri):
    addr: str

    @classmethod
    def from_urlparts(cls, urlparts: ParseResult, uri: str) -> MailUri:
        if urlparts.path.startswith("/"):
            raise ValueError("Authority component not supported: %s" % uri)

        if not urlparts.path:
            raise ValueError("No path component found: %s" % uri)

        return cls(scheme=urlparts.scheme, uri=uri, addr=urlparts.path)


@dataclass
class GeoUri(Uri):
    lat: str
    lon: str
    alt: str

    @classmethod
    def from_urlparts(cls, urlparts: ParseResult, uri: str) -> GeoUri:
        lat, _, lon_alt = urlparts.path.partition(",")
        if not lat or not lon_alt:
            raise ValueError("No latitude or longitude")

        lon, _, alt = lon_alt.partition(",")
        if not lon:
            raise ValueError("No longitude")

        return cls(scheme=urlparts.scheme, uri=uri, lat=lat, lon=lon, alt=alt)


@dataclass
class FileUri(Uri):
    netloc: str
    path: str

    @classmethod
    def from_urlparts(cls, urlparts: ParseResult, uri: str) -> FileUri:
        path = Gio.File.new_for_uri(uri).get_path()
        assert path is not None
        return cls(scheme=urlparts.scheme, uri=uri, netloc=urlparts.netloc, path=path)


@dataclass
class DataUri(Uri):
    media_type: str
    parameters: dict[str, str]
    is_base64: bool
    data: str

    @classmethod
    def from_urlparts(cls, urlparts: ParseResult, uri: str) -> DataUri:
        if urlparts.scheme != "data":
            raise ValueError("Invalid scheme: %s" % urlparts.scheme)

        try:
            metadata, data = urlparts.path.split(",", maxsplit=1)
        except Exception:
            raise ValueError("Invalid data uri: %s" % uri)

        media_type = "text/plain"
        is_base64 = False
        parameters: dict[str, str] = {}

        for index, part in enumerate(metadata.split(";")):
            if not part:
                continue

            if index == 0:
                if "/" in part:
                    media_type = part
                    continue

            if "=" in part:
                key, value = part.split("=")
                parameters[key] = value
                continue

            if part == "base64":
                is_base64 = True
                break

            log.warning("Ignore data uri part '%s', in %s", part, uri)

        return cls(
            scheme="data",
            uri=uri,
            media_type=media_type,
            parameters=parameters,
            is_base64=is_base64,
            data=data,
        )

    @classmethod
    def from_string(cls, uri: str) -> DataUri:
        urlparts = urlparse(uri)
        return DataUri.from_urlparts(urlparts, uri)


UriT = Uri | InvalidUri | XmppIri | MailUri | GeoUri | FileUri | DataUri


SCHEME_CLASS_MAP = {
    "xmpp": XmppIri,
    "mailto": MailUri,
    "geo": GeoUri,
    "file": FileUri,
    "data": DataUri,
}


def is_known_uri_scheme(scheme: str) -> bool:
    """
    `scheme` is lower-case
    """
    if not scheme:
        return False
    if scheme in URI_SCHEMES:
        return True
    if scheme in NONREGISTERED_URI_SCHEMES:
        return True
    return scheme in app.settings.get("additional_uri_schemes").split()


def parse_xmpp_uri_query(pct_iquerycomp: str) -> tuple[str, dict[str, str]]:
    """
    Parses 'mess%61ge;b%6Fdy=Hello%20%F0%9F%8C%90%EF%B8%8F' into
    ('message', {'body': 'Hello 🌐️'}), empty string into ('', {}).
    """
    # Syntax and terminology from
    # <https://rfc-editor.org/rfc/rfc5122#section-2.2>; the pct_ prefix means
    # percent encoding not being undone yet.

    if not pct_iquerycomp:
        return "", {}

    pct_iquerytype, _, pct_ipairs = pct_iquerycomp.partition(";")
    iquerytype = unquote(pct_iquerytype, errors="strict")
    if not pct_ipairs:
        return iquerytype, {}

    pairs: dict[str, str] = {}
    for pct_ipair in pct_ipairs.split(";"):
        pct_ikey, _, pct_ival = pct_ipair.partition("=")
        pairs[unquote(pct_ikey, errors="strict")] = unquote(pct_ival, errors="replace")
    return iquerytype, pairs


def parse_uri(uri: str) -> UriT:
    try:
        urlparts = urlparse(uri)
    except Exception as error:
        return InvalidUri(scheme="", uri=uri, error=str(error))

    scheme = urlparts.scheme

    if not is_known_uri_scheme(scheme):
        return InvalidUri(scheme="", uri=uri, error="Unknown scheme")

    uri_class = SCHEME_CLASS_MAP.get(scheme, Uri)
    try:
        return uri_class.from_urlparts(urlparts, uri)
    except Exception as error:
        return InvalidUri(scheme="", uri=uri, error=str(error))


def filesystem_path_from_uri(uri: str) -> Path | None:
    puri = parse_uri(uri)
    if not isinstance(puri, FileUri):
        log.warning("Not a FileUri: %s", puri)
        return None

    if puri.netloc and puri.netloc.lower() != "localhost" and sys.platform != "win32":
        return None
    return Path(puri.path)


def get_file_path_from_dnd_dropped_uri(text: str) -> Path | None:
    uri = text.strip("\r\n\x00")  # remove \r\n and NULL
    return filesystem_path_from_uri(uri)


def get_file_path_from_uri(uri: str) -> Path | None:
    """Prepare URI by:
    - removing invalid characters
    - parsing and checking the URI for validity
    - checking if the path is a file
    - checking if the file is accessible (path.is_file() calls Path.stat())
    """
    path = get_file_path_from_dnd_dropped_uri(uri)
    if path is None:
        return None

    try:
        is_file = path.is_file()
    except OSError:
        log.exception("Could not access file: %s", path)
        return None

    if not is_file:
        log.warning("Path is not a file: %s", path)
        return None

    return path


def geo_provider_from_location(lat: str, lon: str) -> str:
    return f"https://www.openstreetmap.org/?mlat={lat}&mlon={lon}&zoom=16"


class Coords(NamedTuple):
    location: str
    lat: str
    lon: str


def split_geo_uri(uri: str) -> Coords:
    # Example:
    # geo:37.786971,-122.399677,122.3;CRS=epsg:32718;U=20;mapcolors=abc
    # Assumption is all coordinates are CRS=WGS-84

    # Remove "geo:"
    coords = uri[4:]

    # Remove arguments
    if ";" in coords:
        coords, _ = coords.split(";", maxsplit=1)

    # Split coords
    coords = coords.split(",")
    if len(coords) not in (2, 3):
        raise ValueError("Invalid geo uri: invalid coord count")

    # Remoove coord-c (altitude)
    if len(coords) == 3:
        coords.pop(2)

    lat, lon = coords
    if float(lat) < -90 or float(lat) > 90:
        raise ValueError("Invalid geo_uri: invalid latitude %s" % lat)

    if float(lon) < -180 or float(lon) > 180:
        raise ValueError("Invalid geo_uri: invalid longitude %s" % lon)

    location = ",".join(coords)
    return Coords(location=location, lat=lat, lon=lon)


def get_geo_choords(uri: str) -> Coords | None:
    if not IRI_RX.fullmatch(uri):
        # urlparse removes whitespace (and who knows what else) from URLs,
        # so can't be used for validation.
        return None

    try:
        urlparts = urlparse(uri)
    except Exception:
        return None

    if urlparts.scheme != "geo":
        return None

    try:
        return split_geo_uri(uri)
    except Exception as err:
        log.info("Bad geo URI %s: %s", uri, err)
        return None



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0144: Roster Item Exchange

from __future__ import annotations

import nbxmpp
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.structs import IqProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common import types
from gajim.common.events import RosterItemExchangeEvent
from gajim.common.i18n import _
from gajim.common.modules.base import BaseModule
from gajim.common.util.jid import InvalidFormat
from gajim.common.util.jid import parse_jid


class RosterItemExchange(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(name='iq',
                          callback=self.received_item,
                          typ='set',
                          ns=Namespace.ROSTERX),
            StanzaHandler(name='message',
                          callback=self.received_item,
                          ns=Namespace.ROSTERX,
                          priority=48),
        ]

    def received_item(self,
                      _con: types.NBXMPPClient,
                      stanza: Iq,
                      properties: IqProperties
                      ) -> None:
        # stanza can be a message or a iq

        self._log.info('Received roster items from %s', stanza.getFrom())

        exchange_items_list = {}
        items_list = stanza.getTag(
            'x', namespace=Namespace.ROSTERX).getChildren()
        if items_list is None:
            raise nbxmpp.NodeProcessed

        action = items_list[0].getAttr('action')
        if not action:
            action = 'add'

        for item in items_list:
            try:
                jid = parse_jid(item.getAttr('jid'))
            except InvalidFormat:
                self._log.warning('Invalid JID: %s, ignoring it',
                                  item.getAttr('jid'))
                continue
            name = item.getAttr('name')
            contact = self._get_contact(JID.from_string(jid))
            groups = []
            same_groups = True
            for group in item.getTags('group'):
                groups.append(group.getData())
                # check that all suggested groups are in the groups we have for
                # this contact
                if not contact or group not in contact.groups:
                    same_groups = False
            if contact:
                # check that all groups we have for this contact are in the
                # suggested groups
                for group in contact.groups:
                    if group not in groups:
                        same_groups = False
                if contact.subscription in ('both', 'to') and same_groups:
                    continue
            exchange_items_list[jid] = [name, groups]

        if not exchange_items_list:
            return

        self._log.info('Items: %s', exchange_items_list)

        app.ged.raise_event(RosterItemExchangeEvent(
            client=self._con,
            jid=properties.jid,
            exchange_items_list=exchange_items_list,
            action=action))

    def send_contacts(self,
                      contacts: list[types.ChatContactT],
                      fjid: str,
                      type_: str = 'message'
                      ) -> None:
        if not app.account_is_available(self._account):
            return

        if type_ == 'message':
            if len(contacts) == 1:
                msg = _('Sent contact: "%(jid)s" (%(name)s)') % {
                    'jid': contacts[0].jid,
                    'name': contacts[0].name}
            else:
                msg = _('Sent contacts:')
                for contact in contacts:
                    msg += f'\n "{contact.jid}" ({contact.name})'
            stanza = nbxmpp.Message(to=app.get_jid_without_resource(fjid),
                                    body=msg)
        elif type_ == 'iq':
            stanza = nbxmpp.Iq(to=fjid, typ='set')
        else:
            raise ValueError

        xdata = stanza.addChild(name='x', namespace=Namespace.ROSTERX)
        for contact in contacts:
            name = contact.name
            xdata.addChild(name='item', attrs={'action': 'add',
                                               'jid': str(contact.jid),
                                               'name': name})
            self._log.info('Send contact: %s %s', contact.jid, name)
        self._con.connection.send(stanza)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import Gtk

from gajim.common import app
from gajim.common.helpers import idle_add_once
from gajim.common.i18n import _

from gajim.gtk.builder import get_builder
from gajim.gtk.window import GajimAppWindow


class QuitDialog(GajimAppWindow):
    def __init__(self) -> None:
        GajimAppWindow.__init__(
            self,
            name="QuitDialog",
            title=_("Quit Gajim"),
            transient_for=app.window,
            modal=True,
            add_window_padding=True,
            header_bar=True,
        )

        self._ui = get_builder("quit_dialog.ui")

        self._connect(self._ui.hide_button, "clicked", self._on_button_clicked)
        self._connect(self._ui.minimize_button, "clicked", self._on_button_clicked)
        self._connect(self._ui.quit_button, "clicked", self._on_button_clicked)

        self.set_child(self._ui.box)

    def _on_button_clicked(self, button: Gtk.Button) -> None:
        action = button.get_name()

        if self._ui.remember_checkbutton.get_active():
            app.settings.set("confirm_on_window_delete", False)
            app.settings.set("action_on_close", action)

        if action == "minimize":
            app.window.minimize()
        elif action == "hide":
            app.window.hide_window()
        elif action == "quit":
            # Call with idle because closing the Quit Dialog
            # writes window size to settings and this can only happen
            # when the CoreApplicaton shutdown was not called first
            idle_add_once(app.app.start_shutdown)

        self.close()

    def _cleanup(self):
        pass



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import json
import logging
from collections.abc import Callable

from nbxmpp.const import AnonymityMode
from nbxmpp.structs import MuclumbusItem
from nbxmpp.structs import MuclumbusResult

from gajim.common import app
from gajim.common import types
from gajim.common.file_transfer_manager import FileTransfer
from gajim.common.helpers import determine_proxy
from gajim.common.modules.base import BaseModule

log = logging.getLogger("gajim.c.m.http_muc_search")

EMPTY_RESULT = MuclumbusResult(first=None, last=None, max=None, end=True, items=[])


class HttpMucSearch(BaseModule):
    def __init__(self, client: types.Client) -> None:
        BaseModule.__init__(self, client)

    def search(
        self,
        keywords: list[str],
        last: str | None,
        callback: Callable[[MuclumbusResult], None],
    ) -> None:
        body: dict[str, str | list[str]] = {"keywords": keywords}
        if last is not None:
            body["after"] = last

        app.ftm.http_request(
            "GET",
            app.settings.get("muclumbus_api_http_uri"),
            content_type="application/json",
            input_=json.dumps(body).encode(),
            proxy=determine_proxy(self._account),
            timeout=3,
            user_data=callback,
            callback=self._on_search_result,
        )

    def _on_search_result(self, obj: FileTransfer) -> None:
        callback = obj.get_user_data()
        try:
            result = obj.get_result()
        except Exception as error:
            log.warning("Error while requesting muc search: %s", error)
            callback(EMPTY_RESULT)
            return

        log.info("Received search result: %s", len(result.content))

        try:
            res = self._parse_response(result.content)
        except Exception as error:
            log.error("Unable to parse response: %s", error)
            res = EMPTY_RESULT

        callback(res)

    def _parse_response(self, content: bytes) -> MuclumbusResult:
        response = json.loads(content)
        result = response["result"]
        items = result.get("items")
        if items is None:
            return EMPTY_RESULT

        results: list[MuclumbusItem] = []
        for item in items:
            try:
                anonymity_mode = AnonymityMode(item["anonymity_mode"])
            except (ValueError, KeyError):
                anonymity_mode = AnonymityMode.UNKNOWN

            results.append(
                MuclumbusItem(
                    jid=item["address"],
                    name=item["name"] or "",
                    nusers=str(item["nusers"] or ""),
                    description=item["description"] or "",
                    language=item["language"] or "",
                    is_open=item["is_open"],
                    anonymity_mode=anonymity_mode,
                )
            )

        return MuclumbusResult(
            first=None,
            last=result["last"],
            max=None,
            end=not result["more"],
            items=results,
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0055: Jabber Search

from __future__ import annotations

from typing import Any

import nbxmpp
from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq

from gajim.common import app
from gajim.common import types
from gajim.common.events import SearchFormReceivedEvent
from gajim.common.events import SearchResultReceivedEvent
from gajim.common.modules.base import BaseModule


class Search(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

    def request_search_fields(self, jid: str) -> None:
        self._log.info('Request search fields from %s', jid)
        iq = nbxmpp.Iq(typ='get', to=jid, queryNS=Namespace.SEARCH)
        self._con.connection.SendAndCallForResponse(iq, self._fields_received)

    def _fields_received(self,
                         _nbxmpp_client: types.NBXMPPClient,
                         stanza: Iq
                         ) -> None:
        data = None
        is_dataform = False

        if nbxmpp.isResultNode(stanza):
            self._log.info('Received search fields from %s', stanza.getFrom())
            tag = stanza.getTag('query', namespace=Namespace.SEARCH)
            if tag is None:
                self._log.info('Invalid stanza: %s', stanza)
                return

            data = tag.getTag('x', namespace=Namespace.DATA)
            if data is not None:
                is_dataform = True
            else:
                data = {}
                query = stanza.getQuery()
                for i in query.getChildren():
                    data[i.getName()] = i.getData()
        else:
            self._log.info('Error: %s', stanza.getError())

        app.ged.raise_event(
            SearchFormReceivedEvent(conn=self._con,
                                    is_dataform=is_dataform,
                                    data=data))

    def send_search_form(self,
                         jid: str,
                         form: SimpleDataForm,
                         is_form: bool
                         ) -> None:
        iq = nbxmpp.Iq(typ='set', to=jid, queryNS=Namespace.SEARCH)
        item = iq.setQuery()
        if is_form:
            item.addChild(node=form)
        else:
            for i in form:
                item.setTagData(i, form[i])

        self._con.connection.SendAndCallForResponse(iq, self._received_result)

    def _received_result(self,
                         _nbxmpp_client: types.NBXMPPClient,
                         stanza: Iq
                         ) -> None:
        data = None
        is_dataform = False

        if nbxmpp.isResultNode(stanza):
            self._log.info('Received result from %s', stanza.getFrom())
            tag = stanza.getTag('query', namespace=Namespace.SEARCH)
            if tag is None:
                self._log.info('Invalid stanza: %s', stanza)
                return

            data = tag.getTag('x', namespace=Namespace.DATA)
            if data is not None:
                is_dataform = True
            else:
                data: list[Any] = []
                for item in tag.getTags('item'):
                    # We also show attributes. jid is there
                    field = item.attrs
                    for i in item.getChildren():
                        field[i.getName()] = i.getData()
                    data.append(field)
        else:
            self._log.info('Error: %s', stanza.getError())

        app.ged.raise_event(
            SearchResultReceivedEvent(conn=self._con,
                                      is_dataform=is_dataform,
                                      data=data))



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

ZONES = {
    "Africa/Abidjan": "Africa/Abidjan",
    "Africa/Accra": "Africa/Accra",
    "Africa/Addis_Ababa": "Africa/Addis Ababa",
    "Africa/Algiers": "Africa/Algiers",
    "Africa/Asmara": "Africa/Asmara",
    "Africa/Asmera": "Africa/Asmera",
    "Africa/Bamako": "Africa/Bamako",
    "Africa/Bangui": "Africa/Bangui",
    "Africa/Banjul": "Africa/Banjul",
    "Africa/Bissau": "Africa/Bissau",
    "Africa/Blantyre": "Africa/Blantyre",
    "Africa/Brazzaville": "Africa/Brazzaville",
    "Africa/Bujumbura": "Africa/Bujumbura",
    "Africa/Cairo": "Africa/Cairo",
    "Africa/Casablanca": "Africa/Casablanca",
    "Africa/Ceuta": "Africa/Ceuta",
    "Africa/Conakry": "Africa/Conakry",
    "Africa/Dakar": "Africa/Dakar",
    "Africa/Dar_es_Salaam": "Africa/Dar es Salaam",
    "Africa/Djibouti": "Africa/Djibouti",
    "Africa/Douala": "Africa/Douala",
    "Africa/El_Aaiun": "Africa/El Aaiun",
    "Africa/Freetown": "Africa/Freetown",
    "Africa/Gaborone": "Africa/Gaborone",
    "Africa/Harare": "Africa/Harare",
    "Africa/Johannesburg": "Africa/Johannesburg",
    "Africa/Juba": "Africa/Juba",
    "Africa/Kampala": "Africa/Kampala",
    "Africa/Khartoum": "Africa/Khartoum",
    "Africa/Kigali": "Africa/Kigali",
    "Africa/Kinshasa": "Africa/Kinshasa",
    "Africa/Lagos": "Africa/Lagos",
    "Africa/Libreville": "Africa/Libreville",
    "Africa/Lome": "Africa/Lome",
    "Africa/Luanda": "Africa/Luanda",
    "Africa/Lubumbashi": "Africa/Lubumbashi",
    "Africa/Lusaka": "Africa/Lusaka",
    "Africa/Malabo": "Africa/Malabo",
    "Africa/Maputo": "Africa/Maputo",
    "Africa/Maseru": "Africa/Maseru",
    "Africa/Mbabane": "Africa/Mbabane",
    "Africa/Mogadishu": "Africa/Mogadishu",
    "Africa/Monrovia": "Africa/Monrovia",
    "Africa/Nairobi": "Africa/Nairobi",
    "Africa/Ndjamena": "Africa/Ndjamena",
    "Africa/Niamey": "Africa/Niamey",
    "Africa/Nouakchott": "Africa/Nouakchott",
    "Africa/Ouagadougou": "Africa/Ouagadougou",
    "Africa/Porto-Novo": "Africa/Porto-Novo",
    "Africa/Sao_Tome": "Africa/Sao Tome",
    "Africa/Timbuktu": "Africa/Timbuktu",
    "Africa/Tripoli": "Africa/Tripoli",
    "Africa/Tunis": "Africa/Tunis",
    "Africa/Windhoek": "Africa/Windhoek",
    "America/Adak": "America/Adak",
    "America/Anchorage": "America/Anchorage",
    "America/Anguilla": "America/Anguilla",
    "America/Antigua": "America/Antigua",
    "America/Araguaina": "America/Araguaina",
    "America/Argentina/Buenos_Aires": "America/Argentina/Buenos Aires",
    "America/Argentina/Catamarca": "America/Argentina/Catamarca",
    "America/Argentina/ComodRivadavia": "America/Argentina/ComodRivadavia",
    "America/Argentina/Cordoba": "America/Argentina/Cordoba",
    "America/Argentina/Jujuy": "America/Argentina/Jujuy",
    "America/Argentina/La_Rioja": "America/Argentina/La Rioja",
    "America/Argentina/Mendoza": "America/Argentina/Mendoza",
    "America/Argentina/Rio_Gallegos": "America/Argentina/Rio Gallegos",
    "America/Argentina/Salta": "America/Argentina/Salta",
    "America/Argentina/San_Juan": "America/Argentina/San Juan",
    "America/Argentina/San_Luis": "America/Argentina/San Luis",
    "America/Argentina/Tucuman": "America/Argentina/Tucuman",
    "America/Argentina/Ushuaia": "America/Argentina/Ushuaia",
    "America/Aruba": "America/Aruba",
    "America/Asuncion": "America/Asuncion",
    "America/Atikokan": "America/Atikokan",
    "America/Atka": "America/Atka",
    "America/Bahia": "America/Bahia",
    "America/Bahia_Banderas": "America/Bahia Banderas",
    "America/Barbados": "America/Barbados",
    "America/Belem": "America/Belem",
    "America/Belize": "America/Belize",
    "America/Blanc-Sablon": "America/Blanc-Sablon",
    "America/Boa_Vista": "America/Boa Vista",
    "America/Bogota": "America/Bogota",
    "America/Boise": "America/Boise",
    "America/Buenos_Aires": "America/Buenos Aires",
    "America/Cambridge_Bay": "America/Cambridge Bay",
    "America/Campo_Grande": "America/Campo Grande",
    "America/Cancun": "America/Cancun",
    "America/Caracas": "America/Caracas",
    "America/Catamarca": "America/Catamarca",
    "America/Cayenne": "America/Cayenne",
    "America/Cayman": "America/Cayman",
    "America/Chicago": "America/Chicago",
    "America/Chihuahua": "America/Chihuahua",
    "America/Ciudad_Juarez": "America/Ciudad Juarez",
    "America/Coral_Harbour": "America/Coral Harbour",
    "America/Cordoba": "America/Cordoba",
    "America/Costa_Rica": "America/Costa Rica",
    "America/Coyhaique": "America/Coyhaique",
    "America/Creston": "America/Creston",
    "America/Cuiaba": "America/Cuiaba",
    "America/Curacao": "America/Curacao",
    "America/Danmarkshavn": "America/Danmarkshavn",
    "America/Dawson": "America/Dawson",
    "America/Dawson_Creek": "America/Dawson Creek",
    "America/Denver": "America/Denver",
    "America/Detroit": "America/Detroit",
    "America/Dominica": "America/Dominica",
    "America/Edmonton": "America/Edmonton",
    "America/Eirunepe": "America/Eirunepe",
    "America/El_Salvador": "America/El Salvador",
    "America/Ensenada": "America/Ensenada",
    "America/Fort_Nelson": "America/Fort Nelson",
    "America/Fort_Wayne": "America/Fort Wayne",
    "America/Fortaleza": "America/Fortaleza",
    "America/Glace_Bay": "America/Glace Bay",
    "America/Godthab": "America/Godthab",
    "America/Goose_Bay": "America/Goose Bay",
    "America/Grand_Turk": "America/Grand Turk",
    "America/Grenada": "America/Grenada",
    "America/Guadeloupe": "America/Guadeloupe",
    "America/Guatemala": "America/Guatemala",
    "America/Guayaquil": "America/Guayaquil",
    "America/Guyana": "America/Guyana",
    "America/Halifax": "America/Halifax",
    "America/Havana": "America/Havana",
    "America/Hermosillo": "America/Hermosillo",
    "America/Indiana/Indianapolis": "America/Indiana/Indianapolis",
    "America/Indiana/Knox": "America/Indiana/Knox",
    "America/Indiana/Marengo": "America/Indiana/Marengo",
    "America/Indiana/Petersburg": "America/Indiana/Petersburg",
    "America/Indiana/Tell_City": "America/Indiana/Tell City",
    "America/Indiana/Vevay": "America/Indiana/Vevay",
    "America/Indiana/Vincennes": "America/Indiana/Vincennes",
    "America/Indiana/Winamac": "America/Indiana/Winamac",
    "America/Indianapolis": "America/Indianapolis",
    "America/Inuvik": "America/Inuvik",
    "America/Iqaluit": "America/Iqaluit",
    "America/Jamaica": "America/Jamaica",
    "America/Jujuy": "America/Jujuy",
    "America/Juneau": "America/Juneau",
    "America/Kentucky/Louisville": "America/Kentucky/Louisville",
    "America/Kentucky/Monticello": "America/Kentucky/Monticello",
    "America/Knox_IN": "America/Knox IN",
    "America/Kralendijk": "America/Kralendijk",
    "America/La_Paz": "America/La Paz",
    "America/Lima": "America/Lima",
    "America/Los_Angeles": "America/Los Angeles",
    "America/Louisville": "America/Louisville",
    "America/Lower_Princes": "America/Lower Princes",
    "America/Maceio": "America/Maceio",
    "America/Managua": "America/Managua",
    "America/Manaus": "America/Manaus",
    "America/Marigot": "America/Marigot",
    "America/Martinique": "America/Martinique",
    "America/Matamoros": "America/Matamoros",
    "America/Mazatlan": "America/Mazatlan",
    "America/Mendoza": "America/Mendoza",
    "America/Menominee": "America/Menominee",
    "America/Merida": "America/Merida",
    "America/Metlakatla": "America/Metlakatla",
    "America/Mexico_City": "America/Mexico City",
    "America/Miquelon": "America/Miquelon",
    "America/Moncton": "America/Moncton",
    "America/Monterrey": "America/Monterrey",
    "America/Montevideo": "America/Montevideo",
    "America/Montreal": "America/Montreal",
    "America/Montserrat": "America/Montserrat",
    "America/Nassau": "America/Nassau",
    "America/New_York": "America/New York",
    "America/Nipigon": "America/Nipigon",
    "America/Nome": "America/Nome",
    "America/Noronha": "America/Noronha",
    "America/North_Dakota/Beulah": "America/North Dakota/Beulah",
    "America/North_Dakota/Center": "America/North Dakota/Center",
    "America/North_Dakota/New_Salem": "America/North Dakota/New Salem",
    "America/Nuuk": "America/Nuuk",
    "America/Ojinaga": "America/Ojinaga",
    "America/Panama": "America/Panama",
    "America/Pangnirtung": "America/Pangnirtung",
    "America/Paramaribo": "America/Paramaribo",
    "America/Phoenix": "America/Phoenix",
    "America/Port-au-Prince": "America/Port-au-Prince",
    "America/Port_of_Spain": "America/Port of Spain",
    "America/Porto_Acre": "America/Porto Acre",
    "America/Porto_Velho": "America/Porto Velho",
    "America/Puerto_Rico": "America/Puerto Rico",
    "America/Punta_Arenas": "America/Punta Arenas",
    "America/Rainy_River": "America/Rainy River",
    "America/Rankin_Inlet": "America/Rankin Inlet",
    "America/Recife": "America/Recife",
    "America/Regina": "America/Regina",
    "America/Resolute": "America/Resolute",
    "America/Rio_Branco": "America/Rio Branco",
    "America/Rosario": "America/Rosario",
    "America/Santa_Isabel": "America/Santa Isabel",
    "America/Santarem": "America/Santarem",
    "America/Santiago": "America/Santiago",
    "America/Santo_Domingo": "America/Santo Domingo",
    "America/Sao_Paulo": "America/Sao Paulo",
    "America/Scoresbysund": "America/Scoresbysund",
    "America/Shiprock": "America/Shiprock",
    "America/Sitka": "America/Sitka",
    "America/St_Barthelemy": "America/St Barthelemy",
    "America/St_Johns": "America/St Johns",
    "America/St_Kitts": "America/St Kitts",
    "America/St_Lucia": "America/St Lucia",
    "America/St_Thomas": "America/St Thomas",
    "America/St_Vincent": "America/St Vincent",
    "America/Swift_Current": "America/Swift Current",
    "America/Tegucigalpa": "America/Tegucigalpa",
    "America/Thule": "America/Thule",
    "America/Thunder_Bay": "America/Thunder Bay",
    "America/Tijuana": "America/Tijuana",
    "America/Toronto": "America/Toronto",
    "America/Tortola": "America/Tortola",
    "America/Vancouver": "America/Vancouver",
    "America/Virgin": "America/Virgin",
    "America/Whitehorse": "America/Whitehorse",
    "America/Winnipeg": "America/Winnipeg",
    "America/Yakutat": "America/Yakutat",
    "America/Yellowknife": "America/Yellowknife",
    "Antarctica/Casey": "Antarctica/Casey",
    "Antarctica/Davis": "Antarctica/Davis",
    "Antarctica/DumontDUrville": "Antarctica/DumontDUrville",
    "Antarctica/Macquarie": "Antarctica/Macquarie",
    "Antarctica/Mawson": "Antarctica/Mawson",
    "Antarctica/McMurdo": "Antarctica/McMurdo",
    "Antarctica/Palmer": "Antarctica/Palmer",
    "Antarctica/Rothera": "Antarctica/Rothera",
    "Antarctica/South_Pole": "Antarctica/South Pole",
    "Antarctica/Syowa": "Antarctica/Syowa",
    "Antarctica/Troll": "Antarctica/Troll",
    "Antarctica/Vostok": "Antarctica/Vostok",
    "Arctic/Longyearbyen": "Arctic/Longyearbyen",
    "Asia/Aden": "Asia/Aden",
    "Asia/Almaty": "Asia/Almaty",
    "Asia/Amman": "Asia/Amman",
    "Asia/Anadyr": "Asia/Anadyr",
    "Asia/Aqtau": "Asia/Aqtau",
    "Asia/Aqtobe": "Asia/Aqtobe",
    "Asia/Ashgabat": "Asia/Ashgabat",
    "Asia/Ashkhabad": "Asia/Ashkhabad",
    "Asia/Atyrau": "Asia/Atyrau",
    "Asia/Baghdad": "Asia/Baghdad",
    "Asia/Bahrain": "Asia/Bahrain",
    "Asia/Baku": "Asia/Baku",
    "Asia/Bangkok": "Asia/Bangkok",
    "Asia/Barnaul": "Asia/Barnaul",
    "Asia/Beirut": "Asia/Beirut",
    "Asia/Bishkek": "Asia/Bishkek",
    "Asia/Brunei": "Asia/Brunei",
    "Asia/Calcutta": "Asia/Calcutta",
    "Asia/Chita": "Asia/Chita",
    "Asia/Choibalsan": "Asia/Choibalsan",
    "Asia/Chongqing": "Asia/Chongqing",
    "Asia/Chungking": "Asia/Chungking",
    "Asia/Colombo": "Asia/Colombo",
    "Asia/Dacca": "Asia/Dacca",
    "Asia/Damascus": "Asia/Damascus",
    "Asia/Dhaka": "Asia/Dhaka",
    "Asia/Dili": "Asia/Dili",
    "Asia/Dubai": "Asia/Dubai",
    "Asia/Dushanbe": "Asia/Dushanbe",
    "Asia/Famagusta": "Asia/Famagusta",
    "Asia/Gaza": "Asia/Gaza",
    "Asia/Harbin": "Asia/Harbin",
    "Asia/Hebron": "Asia/Hebron",
    "Asia/Ho_Chi_Minh": "Asia/Ho Chi Minh",
    "Asia/Hong_Kong": "Asia/Hong Kong",
    "Asia/Hovd": "Asia/Hovd",
    "Asia/Irkutsk": "Asia/Irkutsk",
    "Asia/Istanbul": "Asia/Istanbul",
    "Asia/Jakarta": "Asia/Jakarta",
    "Asia/Jayapura": "Asia/Jayapura",
    "Asia/Jerusalem": "Asia/Jerusalem",
    "Asia/Kabul": "Asia/Kabul",
    "Asia/Kamchatka": "Asia/Kamchatka",
    "Asia/Karachi": "Asia/Karachi",
    "Asia/Kashgar": "Asia/Kashgar",
    "Asia/Kathmandu": "Asia/Kathmandu",
    "Asia/Katmandu": "Asia/Katmandu",
    "Asia/Khandyga": "Asia/Khandyga",
    "Asia/Kolkata": "Asia/Kolkata",
    "Asia/Krasnoyarsk": "Asia/Krasnoyarsk",
    "Asia/Kuala_Lumpur": "Asia/Kuala Lumpur",
    "Asia/Kuching": "Asia/Kuching",
    "Asia/Kuwait": "Asia/Kuwait",
    "Asia/Macao": "Asia/Macao",
    "Asia/Macau": "Asia/Macau",
    "Asia/Magadan": "Asia/Magadan",
    "Asia/Makassar": "Asia/Makassar",
    "Asia/Manila": "Asia/Manila",
    "Asia/Muscat": "Asia/Muscat",
    "Asia/Nicosia": "Asia/Nicosia",
    "Asia/Novokuznetsk": "Asia/Novokuznetsk",
    "Asia/Novosibirsk": "Asia/Novosibirsk",
    "Asia/Omsk": "Asia/Omsk",
    "Asia/Oral": "Asia/Oral",
    "Asia/Phnom_Penh": "Asia/Phnom Penh",
    "Asia/Pontianak": "Asia/Pontianak",
    "Asia/Pyongyang": "Asia/Pyongyang",
    "Asia/Qatar": "Asia/Qatar",
    "Asia/Qostanay": "Asia/Qostanay",
    "Asia/Qyzylorda": "Asia/Qyzylorda",
    "Asia/Rangoon": "Asia/Rangoon",
    "Asia/Riyadh": "Asia/Riyadh",
    "Asia/Saigon": "Asia/Saigon",
    "Asia/Sakhalin": "Asia/Sakhalin",
    "Asia/Samarkand": "Asia/Samarkand",
    "Asia/Seoul": "Asia/Seoul",
    "Asia/Shanghai": "Asia/Shanghai",
    "Asia/Singapore": "Asia/Singapore",
    "Asia/Srednekolymsk": "Asia/Srednekolymsk",
    "Asia/Taipei": "Asia/Taipei",
    "Asia/Tashkent": "Asia/Tashkent",
    "Asia/Tbilisi": "Asia/Tbilisi",
    "Asia/Tehran": "Asia/Tehran",
    "Asia/Tel_Aviv": "Asia/Tel Aviv",
    "Asia/Thimbu": "Asia/Thimbu",
    "Asia/Thimphu": "Asia/Thimphu",
    "Asia/Tokyo": "Asia/Tokyo",
    "Asia/Tomsk": "Asia/Tomsk",
    "Asia/Ujung_Pandang": "Asia/Ujung Pandang",
    "Asia/Ulaanbaatar": "Asia/Ulaanbaatar",
    "Asia/Ulan_Bator": "Asia/Ulan Bator",
    "Asia/Urumqi": "Asia/Urumqi",
    "Asia/Ust-Nera": "Asia/Ust-Nera",
    "Asia/Vientiane": "Asia/Vientiane",
    "Asia/Vladivostok": "Asia/Vladivostok",
    "Asia/Yakutsk": "Asia/Yakutsk",
    "Asia/Yangon": "Asia/Yangon",
    "Asia/Yekaterinburg": "Asia/Yekaterinburg",
    "Asia/Yerevan": "Asia/Yerevan",
    "Atlantic/Azores": "Atlantic/Azores",
    "Atlantic/Bermuda": "Atlantic/Bermuda",
    "Atlantic/Canary": "Atlantic/Canary",
    "Atlantic/Cape_Verde": "Atlantic/Cape Verde",
    "Atlantic/Faeroe": "Atlantic/Faeroe",
    "Atlantic/Faroe": "Atlantic/Faroe",
    "Atlantic/Jan_Mayen": "Atlantic/Jan Mayen",
    "Atlantic/Madeira": "Atlantic/Madeira",
    "Atlantic/Reykjavik": "Atlantic/Reykjavik",
    "Atlantic/South_Georgia": "Atlantic/South Georgia",
    "Atlantic/St_Helena": "Atlantic/St Helena",
    "Atlantic/Stanley": "Atlantic/Stanley",
    "Australia/ACT": "Australia/ACT",
    "Australia/Adelaide": "Australia/Adelaide",
    "Australia/Brisbane": "Australia/Brisbane",
    "Australia/Broken_Hill": "Australia/Broken Hill",
    "Australia/Canberra": "Australia/Canberra",
    "Australia/Currie": "Australia/Currie",
    "Australia/Darwin": "Australia/Darwin",
    "Australia/Eucla": "Australia/Eucla",
    "Australia/Hobart": "Australia/Hobart",
    "Australia/LHI": "Australia/LHI",
    "Australia/Lindeman": "Australia/Lindeman",
    "Australia/Lord_Howe": "Australia/Lord Howe",
    "Australia/Melbourne": "Australia/Melbourne",
    "Australia/North": "Australia/North",
    "Australia/NSW": "Australia/NSW",
    "Australia/Perth": "Australia/Perth",
    "Australia/Queensland": "Australia/Queensland",
    "Australia/South": "Australia/South",
    "Australia/Sydney": "Australia/Sydney",
    "Australia/Tasmania": "Australia/Tasmania",
    "Australia/Victoria": "Australia/Victoria",
    "Australia/West": "Australia/West",
    "Australia/Yancowinna": "Australia/Yancowinna",
    "Brazil/Acre": "Brazil/Acre",
    "Brazil/DeNoronha": "Brazil/DeNoronha",
    "Brazil/East": "Brazil/East",
    "Brazil/West": "Brazil/West",
    "Canada/Atlantic": "Canada/Atlantic",
    "Canada/Central": "Canada/Central",
    "Canada/Eastern": "Canada/Eastern",
    "Canada/Mountain": "Canada/Mountain",
    "Canada/Newfoundland": "Canada/Newfoundland",
    "Canada/Pacific": "Canada/Pacific",
    "Canada/Saskatchewan": "Canada/Saskatchewan",
    "Canada/Yukon": "Canada/Yukon",
    "CET": "CET",
    "Chile/Continental": "Chile/Continental",
    "Chile/EasterIsland": "Chile/EasterIsland",
    "CST6CDT": "CST6CDT",
    "Cuba": "Cuba",
    "EET": "EET",
    "Egypt": "Egypt",
    "Eire": "Eire",
    "EST": "EST",
    "EST5EDT": "EST5EDT",
    "Etc/GMT": "Etc/GMT",
    "Etc/GMT+0": "Etc/GMT+0",
    "Etc/GMT+1": "Etc/GMT+1",
    "Etc/GMT+10": "Etc/GMT+10",
    "Etc/GMT+11": "Etc/GMT+11",
    "Etc/GMT+12": "Etc/GMT+12",
    "Etc/GMT+2": "Etc/GMT+2",
    "Etc/GMT+3": "Etc/GMT+3",
    "Etc/GMT+4": "Etc/GMT+4",
    "Etc/GMT+5": "Etc/GMT+5",
    "Etc/GMT+6": "Etc/GMT+6",
    "Etc/GMT+7": "Etc/GMT+7",
    "Etc/GMT+8": "Etc/GMT+8",
    "Etc/GMT+9": "Etc/GMT+9",
    "Etc/GMT-0": "Etc/GMT-0",
    "Etc/GMT-1": "Etc/GMT-1",
    "Etc/GMT-10": "Etc/GMT-10",
    "Etc/GMT-11": "Etc/GMT-11",
    "Etc/GMT-12": "Etc/GMT-12",
    "Etc/GMT-13": "Etc/GMT-13",
    "Etc/GMT-14": "Etc/GMT-14",
    "Etc/GMT-2": "Etc/GMT-2",
    "Etc/GMT-3": "Etc/GMT-3",
    "Etc/GMT-4": "Etc/GMT-4",
    "Etc/GMT-5": "Etc/GMT-5",
    "Etc/GMT-6": "Etc/GMT-6",
    "Etc/GMT-7": "Etc/GMT-7",
    "Etc/GMT-8": "Etc/GMT-8",
    "Etc/GMT-9": "Etc/GMT-9",
    "Etc/GMT0": "Etc/GMT0",
    "Etc/Greenwich": "Etc/Greenwich",
    "Etc/UCT": "Etc/UCT",
    "Etc/Universal": "Etc/Universal",
    "Etc/UTC": "Etc/UTC",
    "Etc/Zulu": "Etc/Zulu",
    "Europe/Amsterdam": "Europe/Amsterdam",
    "Europe/Andorra": "Europe/Andorra",
    "Europe/Astrakhan": "Europe/Astrakhan",
    "Europe/Athens": "Europe/Athens",
    "Europe/Belfast": "Europe/Belfast",
    "Europe/Belgrade": "Europe/Belgrade",
    "Europe/Berlin": "Europe/Berlin",
    "Europe/Bratislava": "Europe/Bratislava",
    "Europe/Brussels": "Europe/Brussels",
    "Europe/Bucharest": "Europe/Bucharest",
    "Europe/Budapest": "Europe/Budapest",
    "Europe/Busingen": "Europe/Busingen",
    "Europe/Chisinau": "Europe/Chisinau",
    "Europe/Copenhagen": "Europe/Copenhagen",
    "Europe/Dublin": "Europe/Dublin",
    "Europe/Gibraltar": "Europe/Gibraltar",
    "Europe/Guernsey": "Europe/Guernsey",
    "Europe/Helsinki": "Europe/Helsinki",
    "Europe/Isle_of_Man": "Europe/Isle of Man",
    "Europe/Istanbul": "Europe/Istanbul",
    "Europe/Jersey": "Europe/Jersey",
    "Europe/Kaliningrad": "Europe/Kaliningrad",
    "Europe/Kiev": "Europe/Kiev",
    "Europe/Kirov": "Europe/Kirov",
    "Europe/Kyiv": "Europe/Kyiv",
    "Europe/Lisbon": "Europe/Lisbon",
    "Europe/Ljubljana": "Europe/Ljubljana",
    "Europe/London": "Europe/London",
    "Europe/Luxembourg": "Europe/Luxembourg",
    "Europe/Madrid": "Europe/Madrid",
    "Europe/Malta": "Europe/Malta",
    "Europe/Mariehamn": "Europe/Mariehamn",
    "Europe/Minsk": "Europe/Minsk",
    "Europe/Monaco": "Europe/Monaco",
    "Europe/Moscow": "Europe/Moscow",
    "Europe/Nicosia": "Europe/Nicosia",
    "Europe/Oslo": "Europe/Oslo",
    "Europe/Paris": "Europe/Paris",
    "Europe/Podgorica": "Europe/Podgorica",
    "Europe/Prague": "Europe/Prague",
    "Europe/Riga": "Europe/Riga",
    "Europe/Rome": "Europe/Rome",
    "Europe/Samara": "Europe/Samara",
    "Europe/San_Marino": "Europe/San Marino",
    "Europe/Sarajevo": "Europe/Sarajevo",
    "Europe/Saratov": "Europe/Saratov",
    "Europe/Simferopol": "Europe/Simferopol",
    "Europe/Skopje": "Europe/Skopje",
    "Europe/Sofia": "Europe/Sofia",
    "Europe/Stockholm": "Europe/Stockholm",
    "Europe/Tallinn": "Europe/Tallinn",
    "Europe/Tirane": "Europe/Tirane",
    "Europe/Tiraspol": "Europe/Tiraspol",
    "Europe/Ulyanovsk": "Europe/Ulyanovsk",
    "Europe/Uzhgorod": "Europe/Uzhgorod",
    "Europe/Vaduz": "Europe/Vaduz",
    "Europe/Vatican": "Europe/Vatican",
    "Europe/Vienna": "Europe/Vienna",
    "Europe/Vilnius": "Europe/Vilnius",
    "Europe/Volgograd": "Europe/Volgograd",
    "Europe/Warsaw": "Europe/Warsaw",
    "Europe/Zagreb": "Europe/Zagreb",
    "Europe/Zaporozhye": "Europe/Zaporozhye",
    "Europe/Zurich": "Europe/Zurich",
    "Factory": "Factory",
    "GB": "GB",
    "GB-Eire": "GB-Eire",
    "GMT": "GMT",
    "GMT+0": "GMT+0",
    "GMT-0": "GMT-0",
    "GMT0": "GMT0",
    "Greenwich": "Greenwich",
    "Hongkong": "Hongkong",
    "HST": "HST",
    "Iceland": "Iceland",
    "Indian/Antananarivo": "Indian/Antananarivo",
    "Indian/Chagos": "Indian/Chagos",
    "Indian/Christmas": "Indian/Christmas",
    "Indian/Cocos": "Indian/Cocos",
    "Indian/Comoro": "Indian/Comoro",
    "Indian/Kerguelen": "Indian/Kerguelen",
    "Indian/Mahe": "Indian/Mahe",
    "Indian/Maldives": "Indian/Maldives",
    "Indian/Mauritius": "Indian/Mauritius",
    "Indian/Mayotte": "Indian/Mayotte",
    "Indian/Reunion": "Indian/Reunion",
    "Iran": "Iran",
    "Israel": "Israel",
    "Jamaica": "Jamaica",
    "Japan": "Japan",
    "Kwajalein": "Kwajalein",
    "Libya": "Libya",
    "MET": "MET",
    "Mexico/BajaNorte": "Mexico/BajaNorte",
    "Mexico/BajaSur": "Mexico/BajaSur",
    "Mexico/General": "Mexico/General",
    "MST": "MST",
    "MST7MDT": "MST7MDT",
    "Navajo": "Navajo",
    "NZ": "NZ",
    "NZ-CHAT": "NZ-CHAT",
    "Pacific/Apia": "Pacific/Apia",
    "Pacific/Auckland": "Pacific/Auckland",
    "Pacific/Bougainville": "Pacific/Bougainville",
    "Pacific/Chatham": "Pacific/Chatham",
    "Pacific/Chuuk": "Pacific/Chuuk",
    "Pacific/Easter": "Pacific/Easter",
    "Pacific/Efate": "Pacific/Efate",
    "Pacific/Enderbury": "Pacific/Enderbury",
    "Pacific/Fakaofo": "Pacific/Fakaofo",
    "Pacific/Fiji": "Pacific/Fiji",
    "Pacific/Funafuti": "Pacific/Funafuti",
    "Pacific/Galapagos": "Pacific/Galapagos",
    "Pacific/Gambier": "Pacific/Gambier",
    "Pacific/Guadalcanal": "Pacific/Guadalcanal",
    "Pacific/Guam": "Pacific/Guam",
    "Pacific/Honolulu": "Pacific/Honolulu",
    "Pacific/Johnston": "Pacific/Johnston",
    "Pacific/Kanton": "Pacific/Kanton",
    "Pacific/Kiritimati": "Pacific/Kiritimati",
    "Pacific/Kosrae": "Pacific/Kosrae",
    "Pacific/Kwajalein": "Pacific/Kwajalein",
    "Pacific/Majuro": "Pacific/Majuro",
    "Pacific/Marquesas": "Pacific/Marquesas",
    "Pacific/Midway": "Pacific/Midway",
    "Pacific/Nauru": "Pacific/Nauru",
    "Pacific/Niue": "Pacific/Niue",
    "Pacific/Norfolk": "Pacific/Norfolk",
    "Pacific/Noumea": "Pacific/Noumea",
    "Pacific/Pago_Pago": "Pacific/Pago Pago",
    "Pacific/Palau": "Pacific/Palau",
    "Pacific/Pitcairn": "Pacific/Pitcairn",
    "Pacific/Pohnpei": "Pacific/Pohnpei",
    "Pacific/Ponape": "Pacific/Ponape",
    "Pacific/Port_Moresby": "Pacific/Port Moresby",
    "Pacific/Rarotonga": "Pacific/Rarotonga",
    "Pacific/Saipan": "Pacific/Saipan",
    "Pacific/Samoa": "Pacific/Samoa",
    "Pacific/Tahiti": "Pacific/Tahiti",
    "Pacific/Tarawa": "Pacific/Tarawa",
    "Pacific/Tongatapu": "Pacific/Tongatapu",
    "Pacific/Truk": "Pacific/Truk",
    "Pacific/Wake": "Pacific/Wake",
    "Pacific/Wallis": "Pacific/Wallis",
    "Pacific/Yap": "Pacific/Yap",
    "Poland": "Poland",
    "Portugal": "Portugal",
    "PRC": "PRC",
    "PST8PDT": "PST8PDT",
    "ROC": "ROC",
    "ROK": "ROK",
    "Singapore": "Singapore",
    "Turkey": "Turkey",
    "UCT": "UCT",
    "Universal": "Universal",
    "US/Alaska": "US/Alaska",
    "US/Aleutian": "US/Aleutian",
    "US/Arizona": "US/Arizona",
    "US/Central": "US/Central",
    "US/East-Indiana": "US/East-Indiana",
    "US/Eastern": "US/Eastern",
    "US/Hawaii": "US/Hawaii",
    "US/Indiana-Starke": "US/Indiana-Starke",
    "US/Michigan": "US/Michigan",
    "US/Mountain": "US/Mountain",
    "US/Pacific": "US/Pacific",
    "US/Samoa": "US/Samoa",
    "UTC": "UTC",
    "W-SU": "W-SU",
    "WET": "WET",
    "Zulu": "Zulu",
}

WINDOWS_ZONES = {
    "AUS Central Standard Time": "Australia/Darwin",
    "AUS Eastern Standard Time": "Australia/Sydney",
    "Afghanistan Standard Time": "Asia/Kabul",
    "Alaskan Standard Time": "America/Anchorage",
    "Aleutian Standard Time": "America/Adak",
    "Altai Standard Time": "Asia/Barnaul",
    "Arab Standard Time": "Asia/Riyadh",
    "Arabian Standard Time": "Asia/Dubai",
    "Arabic Standard Time": "Asia/Baghdad",
    "Argentina Standard Time": "America/Buenos_Aires",
    "Astrakhan Standard Time": "Europe/Astrakhan",
    "Atlantic Standard Time": "America/Halifax",
    "Aus Central W. Standard Time": "Australia/Eucla",
    "Azerbaijan Standard Time": "Asia/Baku",
    "Azores Standard Time": "Atlantic/Azores",
    "Bahia Standard Time": "America/Bahia",
    "Bangladesh Standard Time": "Asia/Dhaka",
    "Belarus Standard Time": "Europe/Minsk",
    "Bougainville Standard Time": "Pacific/Bougainville",
    "Canada Central Standard Time": "America/Regina",
    "Cape Verde Standard Time": "Atlantic/Cape_Verde",
    "Caucasus Standard Time": "Asia/Yerevan",
    "Cen. Australia Standard Time": "Australia/Adelaide",
    "Central America Standard Time": "America/Guatemala",
    "Central Asia Standard Time": "Asia/Almaty",
    "Central Brazilian Standard Time": "America/Cuiaba",
    "Central Europe Standard Time": "Europe/Budapest",
    "Central European Standard Time": "Europe/Warsaw",
    "Central Pacific Standard Time": "Pacific/Guadalcanal",
    "Central Standard Time": "America/Chicago",
    "Central Standard Time (Mexico)": "America/Mexico_City",
    "Chatham Islands Standard Time": "Pacific/Chatham",
    "China Standard Time": "Asia/Shanghai",
    "Cuba Standard Time": "America/Havana",
    "Dateline Standard Time": "Etc/GMT+12",
    "E. Africa Standard Time": "Africa/Nairobi",
    "E. Australia Standard Time": "Australia/Brisbane",
    "E. Europe Standard Time": "Europe/Chisinau",
    "E. South America Standard Time": "America/Sao_Paulo",
    "Easter Island Standard Time": "Pacific/Easter",
    "Eastern Standard Time": "America/New_York",
    "Eastern Standard Time (Mexico)": "America/Cancun",
    "Egypt Standard Time": "Africa/Cairo",
    "Ekaterinburg Standard Time": "Asia/Yekaterinburg",
    "FLE Standard Time": "Europe/Kiev",
    "Fiji Standard Time": "Pacific/Fiji",
    "GMT Standard Time": "Europe/London",
    "GTB Standard Time": "Europe/Bucharest",
    "Georgian Standard Time": "Asia/Tbilisi",
    "Greenland Standard Time": "America/Godthab",
    "Greenwich Standard Time": "Atlantic/Reykjavik",
    "Haiti Standard Time": "America/Port-au-Prince",
    "Hawaiian Standard Time": "Pacific/Honolulu",
    "India Standard Time": "Asia/Calcutta",
    "Iran Standard Time": "Asia/Tehran",
    "Israel Standard Time": "Asia/Jerusalem",
    "Jordan Standard Time": "Asia/Amman",
    "Kaliningrad Standard Time": "Europe/Kaliningrad",
    "Korea Standard Time": "Asia/Seoul",
    "Libya Standard Time": "Africa/Tripoli",
    "Line Islands Standard Time": "Pacific/Kiritimati",
    "Lord Howe Standard Time": "Australia/Lord_Howe",
    "Magadan Standard Time": "Asia/Magadan",
    "Magallanes Standard Time": "America/Punta_Arenas",
    "Marquesas Standard Time": "Pacific/Marquesas",
    "Mauritius Standard Time": "Indian/Mauritius",
    "Middle East Standard Time": "Asia/Beirut",
    "Montevideo Standard Time": "America/Montevideo",
    "Morocco Standard Time": "Africa/Casablanca",
    "Mountain Standard Time": "America/Denver",
    "Mountain Standard Time (Mexico)": "America/Mazatlan",
    "Myanmar Standard Time": "Asia/Rangoon",
    "N. Central Asia Standard Time": "Asia/Novosibirsk",
    "Namibia Standard Time": "Africa/Windhoek",
    "Nepal Standard Time": "Asia/Katmandu",
    "New Zealand Standard Time": "Pacific/Auckland",
    "Newfoundland Standard Time": "America/St_Johns",
    "Norfolk Standard Time": "Pacific/Norfolk",
    "North Asia East Standard Time": "Asia/Irkutsk",
    "North Asia Standard Time": "Asia/Krasnoyarsk",
    "North Korea Standard Time": "Asia/Pyongyang",
    "Omsk Standard Time": "Asia/Omsk",
    "Pacific SA Standard Time": "America/Santiago",
    "Pacific Standard Time": "America/Los_Angeles",
    "Pacific Standard Time (Mexico)": "America/Tijuana",
    "Pakistan Standard Time": "Asia/Karachi",
    "Paraguay Standard Time": "America/Asuncion",
    "Qyzylorda Standard Time": "Asia/Qyzylorda",
    "Romance Standard Time": "Europe/Paris",
    "Russia Time Zone 10": "Asia/Srednekolymsk",
    "Russia Time Zone 11": "Asia/Kamchatka",
    "Russia Time Zone 3": "Europe/Samara",
    "Russian Standard Time": "Europe/Moscow",
    "SA Eastern Standard Time": "America/Cayenne",
    "SA Pacific Standard Time": "America/Bogota",
    "SA Western Standard Time": "America/La_Paz",
    "SE Asia Standard Time": "Asia/Bangkok",
    "Saint Pierre Standard Time": "America/Miquelon",
    "Sakhalin Standard Time": "Asia/Sakhalin",
    "Samoa Standard Time": "Pacific/Apia",
    "Sao Tome Standard Time": "Africa/Sao_Tome",
    "Saratov Standard Time": "Europe/Saratov",
    "Singapore Standard Time": "Asia/Singapore",
    "South Africa Standard Time": "Africa/Johannesburg",
    "South Sudan Standard Time": "Africa/Juba",
    "Sri Lanka Standard Time": "Asia/Colombo",
    "Sudan Standard Time": "Africa/Khartoum",
    "Syria Standard Time": "Asia/Damascus",
    "Taipei Standard Time": "Asia/Taipei",
    "Tasmania Standard Time": "Australia/Hobart",
    "Tocantins Standard Time": "America/Araguaina",
    "Tokyo Standard Time": "Asia/Tokyo",
    "Tomsk Standard Time": "Asia/Tomsk",
    "Tonga Standard Time": "Pacific/Tongatapu",
    "Transbaikal Standard Time": "Asia/Chita",
    "Turkey Standard Time": "Europe/Istanbul",
    "Turks And Caicos Standard Time": "America/Grand_Turk",
    "US Eastern Standard Time": "America/Indianapolis",
    "US Mountain Standard Time": "America/Phoenix",
    "UTC": "Etc/UTC",
    "UTC+12": "Etc/GMT-12",
    "UTC+13": "Etc/GMT-13",
    "UTC-02": "Etc/GMT+2",
    "UTC-08": "Etc/GMT+8",
    "UTC-09": "Etc/GMT+9",
    "UTC-11": "Etc/GMT+11",
    "Ulaanbaatar Standard Time": "Asia/Ulaanbaatar",
    "Venezuela Standard Time": "America/Caracas",
    "Vladivostok Standard Time": "Asia/Vladivostok",
    "Volgograd Standard Time": "Europe/Volgograd",
    "W. Australia Standard Time": "Australia/Perth",
    "W. Central Africa Standard Time": "Africa/Lagos",
    "W. Europe Standard Time": "Europe/Berlin",
    "W. Mongolia Standard Time": "Asia/Hovd",
    "West Asia Standard Time": "Asia/Tashkent",
    "West Bank Standard Time": "Asia/Hebron",
    "West Pacific Standard Time": "Pacific/Port_Moresby",
    "Yakutsk Standard Time": "Asia/Yakutsk",
    "Yukon Standard Time": "America/Whitehorse",
}



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0077: In-Band Registration

from __future__ import annotations

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import DiscoInfo

from gajim.common import types
from gajim.common.modules.base import BaseModule


class Register(BaseModule):

    _nbxmpp_extends = 'Register'
    _nbxmpp_methods = [
        'unregister',
        'change_password',
        'change_password_with_form',
        'request_register_form',
        'submit_register_form',
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.supported = False

    def pass_disco(self, info: DiscoInfo) -> None:
        self.supported = Namespace.REGISTER in info.features



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0363: HTTP File Upload

from __future__ import annotations

from typing import cast

import binascii
import os
from collections import defaultdict
from pathlib import Path
from urllib.parse import urlparse

from gi.repository import GLib
from gi.repository import GObject
from nbxmpp.errors import HTTPUploadStanzaError
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import HTTPUploadData
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import types
from gajim.common.aes import AESKeyData
from gajim.common.const import FTState
from gajim.common.events import FeatureDiscovered
from gajim.common.events import HTTPUploadError
from gajim.common.events import HTTPUploadStarted
from gajim.common.exceptions import FileError
from gajim.common.file_transfer_manager import FileTransfer as FileTransferM
from gajim.common.filetransfer import FileTransfer
from gajim.common.helpers import determine_proxy
from gajim.common.i18n import _
from gajim.common.modules.base import BaseModule
from gajim.common.multiprocess.http import CancelledError
from gajim.common.structs import OutgoingMessage
from gajim.common.util.preview import guess_mime_type


class HTTPUpload(BaseModule):
    _nbxmpp_extends = "HTTPUpload"

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.available = False
        self.component: JID | None = None
        self.httpupload_namespace: str | None = None
        self.max_file_size: float | None = None  # max file size in bytes

        self._requests_in_progress: dict[int, FileTransferM] = {}

        self._running_transfers: dict[tuple[str, JID], set[HTTPFileTransfer]] = (
            defaultdict(set)
        )

    def pass_disco(self, info: DiscoInfo) -> None:
        if not info.has_httpupload:
            return

        self.available = True
        self.httpupload_namespace = Namespace.HTTPUPLOAD_0
        self.component = info.jid
        self.max_file_size = info.httpupload_max_file_size

        self._log.info("Discovered component: %s", info.jid)

        if self.max_file_size is None:
            self._log.warning("Component does not provide maximum file size")
        else:
            size = GLib.format_size_full(
                int(self.max_file_size), GLib.FormatSizeFlags.IEC_UNITS
            )
            self._log.info("Component has a maximum file size of: %s", size)

        app.ged.raise_event(
            FeatureDiscovered(account=self._account, feature=Namespace.HTTPUPLOAD_0)
        )

    def get_running_transfers(
        self, contact: types.ChatContactT
    ) -> set[HTTPFileTransfer] | None:
        return self._running_transfers.get((contact.account, contact.jid))

    def send_file(self, contact: types.ChatContactT, path: Path) -> None:
        encryption = contact.settings.get("encryption") or None

        try:
            transfer = self._make_transfer(path, encryption, contact)
        except FileError as error:
            event = HTTPUploadError(
                contact.account, contact.jid, _("Could not open file (%s)") % str(error)
            )
            app.ged.raise_event(event)
            return

        transfer.connect("cancel", self._on_cancel_upload)
        transfer.connect("state-changed", self._on_http_upload_state_changed)

        event = HTTPUploadStarted(contact.account, contact.jid, transfer)
        app.ged.raise_event(event)

        transfer.set_preparing()
        self._log.info("Sending request for slot")
        self._nbxmpp("HTTPUpload").request_slot(
            jid=self.component,
            filename=transfer.filename,
            size=transfer.size,
            content_type=transfer.mime,
            callback=self._received_slot,
            user_data=transfer,
        )

    def _make_transfer(
        self,
        path: Path,
        encryption: str | None,
        contact: types.ChatContactT,
    ) -> HTTPFileTransfer:
        if not path or not path.exists():
            raise FileError(_("Could not access file"))

        invalid_file = False
        msg = ""
        stat = path.stat()

        if os.path.isfile(path):
            if stat[6] == 0:
                invalid_file = True
                msg = _("File is empty")
        else:
            invalid_file = True
            msg = _("File does not exist")

        if self.max_file_size is not None and stat.st_size > self.max_file_size:
            invalid_file = True
            size = GLib.format_size_full(
                int(self.max_file_size), GLib.FormatSizeFlags.IEC_UNITS
            )
            msg = _("File is too large, maximum allowed file size is: %s") % size

        if invalid_file:
            raise FileError(msg)

        mime = guess_mime_type(path)
        if not mime:
            mime = "application/octet-stream"  # fallback mime type
        self._log.info("Detected MIME type of file: %s", mime)

        transfer = HTTPFileTransfer(
            self._account, str(path), contact, mime, encryption, contact.is_groupchat
        )

        key = (contact.account, contact.jid)
        self._running_transfers[key].add(transfer)

        return transfer

    def _on_http_upload_state_changed(
        self, transfer: HTTPFileTransfer, _signal_name: str, state: FTState
    ) -> None:
        if state.is_finished:
            uri = transfer.get_transformed_uri()

            message = OutgoingMessage(
                account=transfer.account,
                contact=transfer.contact,
                text=uri,
                oob_url=uri,
            )

            self._client.send_message(message)

    def _on_cancel_upload(self, transfer: HTTPFileTransfer, _signal_name: str) -> None:
        transfer.set_cancelled()

        key = (transfer.account, transfer.contact.jid)
        self._running_transfers[key].discard(transfer)

        obj = self._requests_in_progress.get(id(transfer))
        if obj is None:
            return

        obj.cancel()

    @staticmethod
    def _uri_is_acceptable(uri: str | None) -> bool:
        if not uri:
            return False

        parts = urlparse(uri)
        if parts.scheme not in ("http", "https"):
            return False

        # Remove port
        domain = parts.netloc.split(":", maxsplit=1)[0]
        if domain.endswith(".onion"):
            return True

        return parts.scheme != "http"

    def _received_slot(self, task: Task) -> None:
        transfer = cast(HTTPFileTransfer, task.get_user_data())

        try:
            result = cast(HTTPUploadData, task.finish())
        except HTTPUploadStanzaError as error:
            if error.app_condition == "file-too-large":
                size = error.get_max_file_size()
                if size is not None:
                    size_text = GLib.format_size_full(
                        int(size), GLib.FormatSizeFlags.IEC_UNITS
                    )
                else:
                    size_text = "Unknown"

                error_text = (
                    _("File is too large, maximum allowed file size is: %s") % size_text
                )
                transfer.set_error("file-too-large", error_text)
            else:
                transfer.set_error("misc", str(error))
            return

        except Exception as error:
            transfer.set_error("misc", str(error))
            return

        transfer.process_result(result)

        if not self._uri_is_acceptable(transfer.put_uri) or not self._uri_is_acceptable(
            transfer.get_uri
        ):
            transfer.set_error("insecure")
            return

        self._log.info("Uploading file to %s", transfer.put_uri)
        self._log.info("Please download from %s", transfer.get_uri)

        self._upload_file(transfer)

    def _upload_file(self, transfer: HTTPFileTransfer) -> None:
        transfer.set_started()

        assert transfer.put_uri is not None
        obj = app.ftm.http_request(
            "PUT",
            transfer.put_uri,
            headers=transfer.headers,
            content_type=transfer.mime,
            encryption_data=transfer.get_encryption_data(),
            input_=transfer.path,
            with_progress=True,
            proxy=determine_proxy(self._account),
            timeout=30,
            user_data=transfer,
            callback=self._on_finish,
        )
        if obj is None:
            return

        obj.connect("notify::progress", self._on_upload_progress)

        self._requests_in_progress[id(transfer)] = obj

    def _on_finish(self, ftobj: FileTransferM) -> None:
        transfer = cast(HTTPFileTransfer, ftobj.get_user_data())

        self._requests_in_progress.pop(id(transfer), None)

        key = (transfer.account, transfer.contact.jid)
        self._running_transfers[key].discard(transfer)

        try:
            ftobj.raise_for_error()
        except CancelledError:
            self._log.info("Upload cancelled")

        except Exception as error:
            self._log.error(error)
            transfer.set_error("http-response", str(error))

        else:
            transfer.set_finished()
            self._log.info("Upload completed successfully")

    def _on_upload_progress(
        self, ftobj: FileTransferM, _param: GObject.ParamSpec
    ) -> None:
        transfer = cast(HTTPFileTransfer, ftobj.get_user_data())
        transfer.set_progress(ftobj.progress)


class HTTPFileTransfer(FileTransfer):
    _state_descriptions = {
        FTState.ENCRYPTING: _("Encrypting file…"),
        FTState.PREPARING: _("Requesting HTTP File Upload Slot…"),
        FTState.STARTED: _("Uploading via HTTP File Upload…"),
    }

    _errors = {
        "insecure": _("The server returned an insecure transport (HTTP)."),
        "encryption-not-available": _(
            "There is no encryption method available for the chosen encryption."
        ),
        "unknown": _("Unknown error."),
    }

    def __init__(
        self,
        account: str,
        path: str,
        contact: types.ChatContactT,
        mime: str,
        encryption: str | None,
        groupchat: bool,
    ) -> None:
        FileTransfer.__init__(self, account)

        self._path = Path(path)
        self._encryption = encryption
        self._groupchat = groupchat
        self._contact = contact
        self._mime = mime

        self.put_uri: str | None = None
        self.get_uri: str | None = None

        self._data: bytes | None = None
        self._headers: dict[str, str] = {}

        self._aes_key_data = None
        if encryption:
            self._aes_key_data = AESKeyData.init()

    @property
    def size(self) -> int:
        size = self._path.stat().st_size
        if self._encryption:
            return size + 16
        return size

    @size.setter
    def size(self, size: int) -> None:
        # Backwards compatibility with plugins
        pass

    @property
    def mime(self) -> str:
        return self._mime

    @property
    def contact(self) -> types.ChatContactT:
        return self._contact

    @property
    def is_groupchat(self) -> bool:
        return self._groupchat

    @property
    def encryption(self) -> str | None:
        return self._encryption

    @property
    def headers(self) -> dict[str, str]:
        return self._headers

    @property
    def path(self) -> Path:
        return self._path

    def get_encryption_data(self) -> AESKeyData | None:
        return self._aes_key_data

    def get_transformed_uri(self) -> str:
        assert self.get_uri is not None
        if self._aes_key_data is not None:
            fragment = binascii.hexlify(
                self._aes_key_data.iv + self._aes_key_data.key
            ).decode()
            return f"aesgcm{self.get_uri[5:]}#{fragment}"

        return self.get_uri

    @property
    def filename(self) -> str:
        return self._path.name

    def set_error(self, domain: str, text: str = "") -> None:
        if not text:
            text = self._errors.get(domain) or self._errors["unknown"]

        super().set_error(domain, text)

    def process_result(self, result: HTTPUploadData) -> None:
        self.put_uri = result.put_uri
        self.get_uri = result.get_uri
        self._headers = result.headers



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging
import traceback
from io import StringIO

from gi.repository import Adw
from gi.repository import GLib
from gi.repository import Gtk

from gajim.common import app
from gajim.common.events import DBMigrationError
from gajim.common.events import DBMigrationFinished
from gajim.common.events import DBMigrationProgress
from gajim.common.events import DBMigrationStart
from gajim.common.ged import EventHelper
from gajim.common.i18n import _

from gajim.gtk.builder import get_builder
from gajim.gtk.util.classes import SignalManager

log = logging.getLogger("gajim.gtk.db_migration")

# Beware that many services are not initialized in Gajim when this dialog runs
# This dialog should not access any Gajim settings, or inherit from classes
# which access Gajim settings.


class DBMigration(Adw.ApplicationWindow, EventHelper, SignalManager):
    def __init__(
        self,
    ) -> None:
        Adw.ApplicationWindow.__init__(
            self,
            name="DBMigration",
            title=_("Database Migration"),
            default_width=600,
            default_height=300,
            modal=True,
        )
        EventHelper.__init__(self)
        SignalManager.__init__(self)

        self.set_deletable(False)
        self.set_resizable(False)

        self._ui = get_builder("db_migration.ui")
        self.set_content(self._ui.box)

        self._connect(
            self._ui.error_copy_button, "clicked", self._on_error_copy_clicked
        )
        self._connect(
            self._ui.error_close_button, "clicked", self._on_close_button_clicked
        )
        self._connect(
            self._ui.success_close_button, "clicked", self._on_close_button_clicked
        )
        self._connect(self, "close-request", self._on_close_request)

        self.register_events(
            [
                ("db-migration-start", 0, self._on_start),
                ("db-migration-progress", 0, self._on_progress),
                ("db-migration-error", 0, self._on_error),
                ("db-migration-finished", 0, self._on_finished),
            ]
        )

        self._timeout_id = GLib.timeout_add(100, self._set_transient)

    def _set_transient(self) -> bool:
        # Set transient on every update to make sure transient is set as
        # soon as main window is available
        self.set_transient_for(app.window)
        return GLib.SOURCE_CONTINUE

    def _on_close_request(self, _widget: Adw.ApplicationWindow) -> None:
        log.debug("Initiate Cleanup: %s", self.get_name())
        self._disconnect_all()
        self.unregister_events()
        GLib.source_remove(self._timeout_id)
        app.check_finalize(self)

    def _on_start(self, event: DBMigrationStart) -> None:
        self._ui.stack.set_visible_child_name("progress-page")
        self._ui.status_label.set_text("0 %")
        self._ui.version_label.set_text(_("Migration to version %s") % event.version)

        context = GLib.MainContext.default()
        while context.pending():
            context.iteration(may_block=False)

    def _on_progress(self, event: DBMigrationProgress) -> None:
        self._ui.status_label.set_text(f"{event.value} %")

        context = GLib.MainContext.default()
        while context.pending():
            context.iteration(may_block=False)

    def _on_finished(self, _event: DBMigrationFinished) -> None:
        self._ui.stack.set_visible_child_name("success-page")

    def _on_error(self, event: DBMigrationError) -> None:
        trace = StringIO()
        traceback.print_exception(
            type(event.exception),
            event.exception,
            event.exception.__traceback__,
            None,
            trace,
        )
        self._ui.error_view.get_buffer().set_text(trace.getvalue())
        self._ui.stack.set_visible_child_name("error-page")

    def _on_error_copy_clicked(self, _button: Gtk.Button) -> None:
        text_buffer = self._ui.error_view.get_buffer()
        start, end = text_buffer.get_bounds()
        error_text = text_buffer.get_text(start, end, True)
        self.get_clipboard().set(error_text)

    def _on_close_button_clicked(self, _button: Gtk.Button) -> None:
        self.close()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import overload

import logging
import multiprocessing
import queue
import threading
from collections import defaultdict
from collections.abc import Callable
from collections.abc import Iterable
from concurrent.futures import Future
from functools import partial
from pathlib import Path
from urllib.parse import urlparse

from gi.repository import GLib
from gi.repository import GObject
from nbxmpp.structs import ProxyData

from gajim.common import app
from gajim.common.aes import AESKeyData
from gajim.common.enum import FTState
from gajim.common.helpers import determine_proxy
from gajim.common.helpers import get_uuid
from gajim.common.multiprocess.http import DEFAULT_MAX_CONTENT_LENGTH
from gajim.common.multiprocess.http import http_request
from gajim.common.multiprocess.http import HTTPResult
from gajim.common.multiprocess.http import TransferMetadata
from gajim.common.multiprocess.http import TransferState
from gajim.common.util.http import get_aes_key_data

log = logging.getLogger("gajim.c.ftm")

QueueT = queue.Queue[TransferState | TransferMetadata]


class FileTransferManager:
    def __init__(self) -> None:
        self._transfers: dict[str, FileTransfer] = {}
        mp_context = multiprocessing.get_context("spawn")
        self._manager = mp_context.Manager()
        self._queue: QueueT = self._manager.Queue()
        GLib.timeout_add(100, self._poll_queue)

    def get_transfer(self, id_: str) -> FileTransfer | None:
        return self._transfers.get(id_)

    def _poll_queue(self) -> bool:
        messages: dict[str, list[TransferState | TransferMetadata]] = defaultdict(list)
        while not self._queue.empty():
            try:
                state = self._queue.get_nowait()
            except Exception:
                pass

            else:
                messages[state.id].append(state)

        for transfer_id, states in messages.items():
            obj = self._transfers.get(transfer_id)
            if obj is None:
                log.error("Unable to find transfer object with id: %s", transfer_id)

            else:
                try:
                    obj.process_multiple_states(states)
                except Exception:
                    log.exception("Failed to process states")

        return GLib.SOURCE_CONTINUE

    def http_request(
        self,
        method: Literal["GET", "POST", "PUT"],
        url: str,
        id_: str | None = None,
        headers: dict[str, str] | None = None,
        params: dict[str, Any] | None = None,
        content_type: str | None = None,
        encryption_data: AESKeyData | None = None,
        input_: Path | bytes = b"",
        output: Path | None = None,
        with_progress: bool = False,
        max_content_length: int = DEFAULT_MAX_CONTENT_LENGTH,
        max_download_size: int = 0,
        allowed_content_types: Iterable[str] | None = None,
        hash_algo: str = "sha256",
        hash_value: str | None = None,
        proxy: ProxyData | None = None,
        timeout: int = 5,
        user_data: Any = None,
        callback: Callable[[FileTransfer], Any] | None = None,
    ) -> FileTransfer | None:
        if id_ is None:
            id_ = get_uuid()

        obj = self._transfers.get(id_)
        if obj is not None:
            return obj

        if proxy is None:
            proxy = determine_proxy()

        urlparts = urlparse(url)

        decryption_data = None
        if urlparts.scheme == "aesgcm":
            try:
                decryption_data = get_aes_key_data(urlparts.fragment)
            except ValueError as error:
                log.exception(error)

            # Don’t send fragment to the server, it would leak the AES key
            urlparts = urlparts._replace(scheme="https", fragment="")

        event = self._manager.Event()

        try:
            future = app.process_pool.submit(
                http_request,
                event,
                id_,
                method,
                urlparts.geturl(),
                timeout,
                queue=self._queue,
                headers=headers,
                params=params,
                content_type=content_type,
                input_=input_,
                output=output,
                with_resp_progress=with_progress and method in ("GET", "POST"),
                with_req_progress=with_progress and method == "PUT",
                max_content_length=max_content_length,
                max_download_size=max_download_size,
                allowed_content_types=allowed_content_types,
                hash_algo=hash_algo,
                hash_value=hash_value,
                encryption_data=encryption_data,
                decryption_data=decryption_data,
                proxy=None if proxy is None else proxy.get_uri(),
            )
        except Exception as error:
            log.exception(error)
            return None

        future.add_done_callback(
            partial(
                GLib.idle_add,
                self._on_http_result,
                id_,
            )
        )

        obj = FileTransfer(self, id_, event, output=output, user_data=user_data)
        if callback is not None:
            obj.connect("finished", callback)
        self._transfers[id_] = obj
        return obj

    def _on_http_result(self, id_: str, future: Future[HTTPResult]) -> None:
        obj = self._transfers.get(id_)
        if obj is None:
            log.error("Unable to find transfer object with id: %s", id_)
            return

        self._poll_queue()

        try:
            download_result = future.result()
        except Exception as error:
            output = obj.get_output_path()
            if output is not None:
                output.unlink(missing_ok=True)
            obj.set_exception(error)

        else:
            obj.set_result(download_result)
            obj.set_finished()

        del self._transfers[id_]


class FileTransfer(GObject.Object):
    __gtype_name__ = "FileTransfer"

    __gsignals__ = {
        "finished": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(
        self,
        manager: FileTransferManager,
        id_: str,
        event: threading.Event,
        output: Path | None = None,
        user_data: Any = None,
    ) -> None:
        super().__init__()

        self._id = id_
        self._state = FTState.CREATED
        self._progress: int = 0
        self._total: int | None = None

        self._event = event
        self._output = output
        self._manager = manager
        self._result = None
        self._exception = None
        self._metadata = None
        self._user_data = user_data

    def cancel(self) -> None:
        self._event.set()

    @GObject.Property(type=str, flags=GObject.ParamFlags.READABLE)
    def id(self) -> str:
        return self._id

    @GObject.Property(type=int, flags=GObject.ParamFlags.READABLE)
    def state(self) -> int:
        return self._state

    @GObject.Property(type=int, flags=GObject.ParamFlags.READABLE)
    def progress(self) -> int:
        return self._progress

    def get_total(self) -> int | None:
        return self._total

    def process_multiple_states(
        self, states: list[TransferState | TransferMetadata]
    ) -> None:
        # Process multiple state changes and notify only once for
        # each changed property

        old_state = self._state
        old_progress = self._progress

        for state in states:
            match state:
                case TransferMetadata():
                    self._metadata = state
                case TransferState():
                    self._state = state.state
                    if self._state == FTState.IN_PROGRESS:
                        self._total = state.total
                        self._progress = state.progress

        if self._state != old_state:
            self.notify("state")
        if self._progress != old_progress:
            self.notify("progress")

    def set_finished(self) -> None:
        self._state = FTState.FINISHED
        self.notify("state")
        self.emit("finished")
        self.run_dispose()

    def is_finished(self) -> bool:
        return self._state == FTState.FINISHED

    def set_user_data(self, user_data: Any) -> None:
        self._user_data = user_data

    def get_user_data(self) -> Any:
        return self._user_data

    def get_metadata(self) -> TransferMetadata | None:
        return self._metadata

    def set_result(self, result: HTTPResult) -> None:
        self._result = result

    @overload
    def get_result(self, *, raise_if_empty: Literal[False]) -> HTTPResult | None: ...

    @overload
    def get_result(self, *, raise_if_empty: Literal[True]) -> HTTPResult: ...

    @overload
    def get_result(self) -> HTTPResult: ...

    def get_result(self, *, raise_if_empty: bool = True) -> HTTPResult | None:
        self.raise_for_error()
        if raise_if_empty:
            if self._result is None or not self._result.content:
                raise ValueError("No result available")

        return self._result

    def set_exception(self, exception: Exception | None) -> None:
        self._exception = exception
        self._state = FTState.ERROR
        self.notify("state")
        self.emit("finished")
        self.run_dispose()

    def raise_for_error(self) -> None:
        if self._exception is not None:
            raise self._exception

    def get_output_path(self) -> Path | None:
        return self._output



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing
from typing import Any
from typing import Union

import datetime
import json
from collections.abc import Callable
from dataclasses import dataclass
from dataclasses import field
from functools import cached_property

from nbxmpp.const import Affiliation
from nbxmpp.const import InviteType
from nbxmpp.const import Role
from nbxmpp.const import StatusCode
from nbxmpp.modules.security_labels import Catalog
from nbxmpp.modules.vcard4 import VCard
from nbxmpp.protocol import JID
from nbxmpp.structs import HTTPAuthData
from nbxmpp.structs import LocationData
from nbxmpp.structs import MAMData
from nbxmpp.structs import RosterItem
from nbxmpp.structs import TuneData

from gajim.common import app
from gajim.common.const import EncryptionInfoMsg
from gajim.common.const import JingleState
from gajim.common.file_props import FileProp
from gajim.common.helpers import get_uuid
from gajim.common.storage.archive import models as mod
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageType
from gajim.common.storage.base import Encoder

if typing.TYPE_CHECKING:
    from gajim.common.client import Client
    from gajim.common.modules.httpupload import HTTPFileTransfer
    from gajim.plugins.manifest import PluginManifest

ChatListEventT = Union[
    'MessageReceived',
    'MessageCorrected',
    'MessageModerated',
    'MessageRetracted',
    'PresenceReceived',
    'MessageSent',
    'MessageDeleted',
    'JingleRequestReceived',
    'FileRequestReceivedEvent'
]


@dataclass
class ApplicationEvent:
    name: str
    context_id: str = field(init=False, default_factory=get_uuid)

    def serialize(self) -> str:
        '''Convert to a json formatted string for database storage'''
        event_data = self.__dict__.copy()

        # Drop timestamp, since it is already handled while storing
        event_data.pop("timestamp", None)
        return json.dumps(event_data, cls=Encoder)


@dataclass
class StyleChanged(ApplicationEvent):
    name: str = field(init=False, default='style-changed')


@dataclass
class ThemeUpdate(ApplicationEvent):
    name: str = field(init=False, default='theme-update')


@dataclass
class ShowChanged(ApplicationEvent):
    name: str = field(init=False, default='our-show')
    account: str
    show: str


@dataclass
class RegisterActions(ApplicationEvent):
    name: str = field(init=False, default='register-actions')


@dataclass
class AccountConnected(ApplicationEvent):
    name: str = field(init=False, default='account-connected')
    account: str


@dataclass
class AccountDisconnected(ApplicationEvent):
    name: str = field(init=False, default='account-disconnected')
    account: str


@dataclass
class PlainConnection(ApplicationEvent):
    name: str = field(init=False, default='plain-connection')
    account: str
    connect: Callable[..., Any]
    abort: Callable[..., Any]


@dataclass
class PasswordRequired(ApplicationEvent):
    name: str = field(init=False, default='password-required')
    client: 'Client'
    on_password: Callable[..., Any]


@dataclass
class Notification(ApplicationEvent):
    name: str = field(init=False, default='notification')
    context_id: str
    account: str
    type: str
    title: str
    text: str
    jid: JID | str | None = None
    sub_type: str | None = None
    sound: str | None = None
    resource: str | None = None


@dataclass
class NotificationWithdrawn(ApplicationEvent):
    name: str = field(init=False, default='notification-withdrawn')
    context_id: str


@dataclass
class ChatRead(ApplicationEvent):
    name: str = field(init=False, default='chat-read')
    account: str
    jid: JID


@dataclass
class StanzaSent(ApplicationEvent):
    name: str = field(init=False, default='stanza-sent')
    account: str
    stanza: Any


@dataclass
class StanzaReceived(ApplicationEvent):
    name: str = field(init=False, default='stanza-received')
    account: str
    stanza: Any


@dataclass
class SignedIn(ApplicationEvent):
    name: str = field(init=False, default='signed-in')
    account: str
    conn: 'Client'


@dataclass
class LocationChanged(ApplicationEvent):
    name: str = field(init=False, default='location-changed')
    info: LocationData | None


@dataclass
class MusicTrackChanged(ApplicationEvent):
    name: str = field(init=False, default='music-track-changed')
    info: TuneData | None


@dataclass
class MessageSent(ApplicationEvent):
    name: str = field(init=False, default='message-sent')
    account: str
    jid: JID
    pk: int
    play_sound: bool = False

    @cached_property
    def message(self) -> mod.Message:
        m = app.storage.archive.get_message_with_pk(self.pk)
        assert m is not None
        return m


@dataclass
class MessageNotSent(ApplicationEvent):
    name: str = field(init=False, default='message-not-sent')
    client: 'Client'
    jid: str
    message: str
    error: str
    time: float


@dataclass
class MessageDeleted(ApplicationEvent):
    name: str = field(init=False, default='message-deleted')
    account: str
    jid: JID
    pk: int


@dataclass
class MessageAcknowledged(ApplicationEvent):
    name: str = field(init=False, default='message-acknowledged')
    account: str
    jid: JID
    pk: int
    stanza_id: str | None


@dataclass
class FileProgress(ApplicationEvent):
    name: str = field(init=False, default='file-progress')
    file_props: FileProp


@dataclass
class FileCompleted(ApplicationEvent):
    name: str = field(init=False, default='file-completed')
    account: str
    file_props: FileProp
    jid: str


@dataclass
class FileError(ApplicationEvent):
    name: str = field(init=False, default='file-error')
    account: str
    file_props: FileProp
    jid: str


@dataclass
class FileHashError(ApplicationEvent):
    name: str = field(init=False, default='file-hash-error')
    account: str
    file_props: FileProp
    jid: str


@dataclass
class FileRequestSent(ApplicationEvent):
    name: str = field(init=False, default='file-request-sent')
    account: str
    file_props: FileProp
    jid: JID


@dataclass
class FileRequestError(ApplicationEvent):
    name: str = field(init=False, default='file-request-error')
    conn: 'Client'
    file_props: FileProp
    jid: str
    error_msg: str = ''


@dataclass
class FileSendError(ApplicationEvent):
    name: str = field(init=False, default='file-send-error')
    account: str
    file_props: FileProp
    jid: str
    error_msg: str = ''


@dataclass
class HTTPUploadStarted(ApplicationEvent):
    name: str = field(init=False, default='http-upload-started')
    account: str
    jid: JID
    transfer: HTTPFileTransfer


@dataclass
class HTTPUploadError(ApplicationEvent):
    name: str = field(init=False, default='http-upload-error')
    account: str
    jid: JID
    error_msg: str


@dataclass
class AccountEnabled(ApplicationEvent):
    name: str = field(init=False, default='account-enabled')
    account: str


@dataclass
class AccountDisabled(ApplicationEvent):
    name: str = field(init=False, default='account-disabled')
    account: str


@dataclass
class AccountCreated(ApplicationEvent):
    name: str = field(init=False, default='account-created')
    account: str


@dataclass
class AccountRemoved(ApplicationEvent):
    name: str = field(init=False, default='account-removed')
    account: str


@dataclass
class FeatureDiscovered(ApplicationEvent):
    name: str = field(init=False, default='feature-discovered')
    account: str
    feature: str


@dataclass
class BookmarksReceived(ApplicationEvent):
    name: str = field(init=False, default='bookmarks-received')
    account: str


@dataclass
class VCard4Received(ApplicationEvent):
    name: str = field(init=False, default='vcard4-received')
    account: str
    vcard: VCard


@dataclass
class TimezoneChanged(ApplicationEvent):
    name: str = field(init=False, default='timezone-changed')
    account: str
    vcard: str | None
    local: str | None


@dataclass
class MucUserBlockChanged(ApplicationEvent):
    name: str = field(init=False, default='muc-user-block-changed')
    account: str
    jid: JID | None = None


@dataclass
class ReadStateSync(ApplicationEvent):
    name: str = field(init=False, default='read-state-sync')
    account: str
    jid: JID
    marker_id: str


@dataclass
class DisplayedReceived(ApplicationEvent):
    name: str = field(init=False, default='displayed-received')
    account: str
    jid: JID
    pk: int

    @cached_property
    def marker(self) -> mod.DisplayedMarker:
        d = app.storage.archive.get_display_marker_with_pk(self.pk)
        assert d is not None
        return d


@dataclass
class ReactionUpdated(ApplicationEvent):
    name: str = field(init=False, default='reaction-updated')
    account: str
    jid: JID
    id: str
    direction: ChatDirection
    occupant: mod.Occupant | None
    emojis: set[str] | None
    message: mod.Message | None
    is_mam_message: bool


@dataclass
class HttpAuth(ApplicationEvent):
    name: str = field(init=False, default='http-auth')
    client: 'Client'
    data: HTTPAuthData
    stanza: Any


@dataclass
class AgentRemoved(ApplicationEvent):
    name: str = field(init=False, default='agent-removed')
    conn: 'Client'
    agent: str
    jid_list: list[str]


@dataclass
class GatewayPromptReceived(ApplicationEvent):
    name: str = field(init=False, default='gateway-prompt-received')
    conn: 'Client'
    fjid: str
    jid: str
    resource: str
    desc: str
    prompt: str
    prompt_jid: str
    stanza: Any


@dataclass
class ServerDiscoReceived(ApplicationEvent):
    name: str = field(init=False, default='server-disco-received')
    account: str


@dataclass
class MucDiscoUpdate(ApplicationEvent):
    name: str = field(init=False, default='muc-disco-update')
    account: str
    jid: JID


@dataclass
class RawMamMessageReceived(ApplicationEvent):
    name: str = field(init=False, default='raw-mam-message-received')
    account: str
    stanza: Any
    properties: Any


@dataclass
class ArchivingIntervalFinished(ApplicationEvent):
    name: str = field(init=False, default='archiving-interval-finished')
    account: str
    query_id: str


@dataclass
class MessageCorrected(ApplicationEvent):
    name: str = field(init=False, default='message-corrected')
    account: str
    jid: JID
    pk: int
    correction_id: str

    @cached_property
    def message(self) -> mod.Message:
        m = app.storage.archive.get_message_with_pk(self.pk)
        assert m is not None
        return m


@dataclass
class MessageModerated(ApplicationEvent):
    name: str = field(init=False, default='message-moderated')
    account: str
    jid: JID
    moderation: mod.Moderation


@dataclass
class MessageRetracted(ApplicationEvent):
    name: str = field(init=False, default='message-retracted')
    account: str
    jid: JID
    retraction: mod.Retraction


@dataclass
class MessageReceived(ApplicationEvent):
    name: str = field(init=False, default='message-received')
    account: str
    jid: JID
    m_type: MessageType
    mam: MAMData | None
    pk: int

    @cached_property
    def message(self) -> mod.Message:
        m = app.storage.archive.get_message_with_pk(self.pk)
        assert m is not None
        return m

    @property
    def from_mam(self) -> bool:
        return self.mam is not None


@dataclass
class MessageError(ApplicationEvent):
    name: str = field(init=False, default='message-error')
    account: str
    jid: JID
    message_id: str
    error: Any


@dataclass
class RosterItemExchangeEvent(ApplicationEvent):
    name: str = field(init=False, default='roster-item-exchange')
    client: 'Client'
    jid: JID
    exchange_items_list: dict[str, list[str]]
    action: str


@dataclass
class RosterReceived(ApplicationEvent):
    name: str = field(init=False, default='roster-received')
    account: str


@dataclass
class RosterPush(ApplicationEvent):
    name: str = field(init=False, default='roster-push')
    account: str
    item: RosterItem


@dataclass
class SearchFormReceivedEvent(ApplicationEvent):
    name: str = field(init=False, default='search-form-received')
    conn: 'Client'
    is_dataform: bool
    data: Any


@dataclass
class SearchResultReceivedEvent(ApplicationEvent):
    name: str = field(init=False, default='search-result-received')
    conn: 'Client'
    is_dataform: bool
    data: Any


@dataclass
class ReceiptReceived(ApplicationEvent):
    name: str = field(init=False, default='receipt-received')
    account: str
    jid: JID
    receipt_id: str


@dataclass
class CallStarted(ApplicationEvent):
    name: str = field(init=False, default='call-started')
    account: str
    resource_jid: JID


@dataclass
class CallStopped(ApplicationEvent):
    name: str = field(init=False, default='call-stopped')
    account: str
    jid: JID


@dataclass
class CallUpdated(ApplicationEvent):
    name: str = field(init=False, default='call-updated')
    jingle_type: str
    audio_state: JingleState
    video_state: JingleState
    audio_sid: str | None
    video_sid: str | None


@dataclass
class JingleEvent(ApplicationEvent):
    name: str
    conn: 'Client'
    account: str
    fjid: str
    jid: JID
    sid: str
    resource: str
    jingle_session: Any


@dataclass
class JingleConnectedReceived(JingleEvent):
    name: str = field(init=False, default='jingle-connected-received')
    media: Any


@dataclass
class JingleDisconnectedReceived(JingleEvent):
    name: str = field(init=False, default='jingle-disconnected-received')
    media: Any
    reason: str


@dataclass
class JingleRequestReceived(JingleEvent):
    name: str = field(init=False, default='jingle-request-received')
    contents: Any


@dataclass
class JingleFtCancelledReceived(JingleEvent):
    name: str = field(init=False, default='jingle-ft-cancelled-received')
    media: Any
    reason: str


@dataclass
class JingleErrorReceived(JingleEvent):
    name: str = field(init=False, default='jingle-error-received')
    reason: str


@dataclass
class MucAdded(ApplicationEvent):
    name: str = field(init=False, default='muc-added')
    account: str
    jid: JID
    select_chat: bool


@dataclass
class MucDecline(ApplicationEvent):
    name: str = field(init=False, default='muc-decline')
    account: str
    muc: JID
    from_: JID
    reason: str | None


@dataclass
class MucInvitation(ApplicationEvent):
    name: str = field(init=False, default='muc-invitation')
    account: str
    info: Any
    muc: JID
    from_: JID
    reason: str | None
    password: str | None
    type: InviteType
    continued: bool
    thread: str | None


@dataclass
class PingSent(ApplicationEvent):
    name: str = field(init=False, default='ping-sent')
    account: str
    contact: Any


@dataclass
class PingError(ApplicationEvent):
    name: str = field(init=False, default='ping-error')
    account: str
    contact: Any
    error: str


@dataclass
class PingReply(ApplicationEvent):
    name: str = field(init=False, default='ping-reply')
    account: str
    contact: Any
    seconds: float


@dataclass
class SecCatalogReceived(ApplicationEvent):
    name: str = field(init=False, default='sec-catalog-received')
    account: str
    jid: JID
    catalog: Catalog


@dataclass
class PresenceReceived(ApplicationEvent):
    name: str = field(init=False, default='presence-received')
    account: str
    conn: 'Client'
    stanza: Any
    prio: int
    need_add_in_roster: bool
    popup: bool
    ptype: str
    jid: JID
    resource: str
    id_: str
    fjid: str
    timestamp: float
    avatar_sha: str | None
    user_nick: str | None
    show: str
    new_show: str
    old_show: str
    status: str
    contact_list: list[str]
    contact: Any


@dataclass
class SubscribePresenceReceived(ApplicationEvent):
    name: str = field(init=False, default='subscription-request')
    conn: 'Client'
    account: str
    jid: str
    fjid: str
    status: str
    user_nick: str
    is_transport: bool


@dataclass
class SubscribedPresenceReceived(ApplicationEvent):
    name: str = field(init=False, default='subscribed-presence-received')
    account: str
    jid: str


@dataclass
class UnsubscribedPresenceReceived(ApplicationEvent):
    name: str = field(init=False, default='unsubscribed-presence-received')
    conn: 'Client'
    account: str
    jid: str


@dataclass
class FileRequestReceivedEvent(ApplicationEvent):
    name: str = field(init=False, default='file-request-received')
    conn: 'Client'
    stanza: Any
    id_: str
    fjid: str
    account: str
    jid: JID
    file_props: FileProp


@dataclass
class AllowGajimUpdateCheck(ApplicationEvent):
    name: str = field(init=False, default='allow-gajim-update-check')


@dataclass
class PluginUpdatesAvailable(ApplicationEvent):
    name: str = field(init=False, default='plugin-updates-available')
    manifests: list[PluginManifest]


@dataclass
class GajimUpdateAvailable(ApplicationEvent):
    name: str = field(init=False, default='gajim-update-available')
    version: str
    setup_url: str


@dataclass
class MUCNicknameChanged(ApplicationEvent):
    name: str = field(init=False, default='muc-nickname-changed')
    is_self: bool
    new_name: str
    old_name: str
    timestamp: datetime.datetime


@dataclass
class MUCRoomConfigChanged(ApplicationEvent):
    name: str = field(init=False, default='muc-room-config-changed')
    timestamp: datetime.datetime
    status_codes: set[StatusCode]


@dataclass
class MUCRoomConfigFinished(ApplicationEvent):
    name: str = field(init=False, default='muc-room-config-finished')
    timestamp: datetime.datetime


@dataclass
class MUCRoomPresenceError(ApplicationEvent):
    name: str = field(init=False, default='muc-room-presence-error')
    timestamp: datetime.datetime
    error: str


@dataclass
class MUCRoomKicked(ApplicationEvent):
    name: str = field(init=False, default='muc-room-kicked')
    timestamp: datetime.datetime
    status_codes: set[StatusCode] | None
    reason: str | None
    actor: str | None


@dataclass
class MUCRoomDestroyed(ApplicationEvent):
    name: str = field(init=False, default='muc-room-destroyed')
    timestamp: datetime.datetime
    reason: str | None
    alternate: JID | None


@dataclass
class MUCRoomVoiceRequestError(ApplicationEvent):
    name: str = field(init=False, default='muc-room-voice-request-error')
    timestamp: datetime.datetime
    message_id: str
    error: Any


@dataclass
class MUCUserHatsChanged(ApplicationEvent):
    name: str = field(init=False, default='muc-user-hats-changed')
    timestamp: datetime.datetime
    is_self: bool
    nick: str
    hats: list[str] | None


@dataclass
class MUCUserJoined(ApplicationEvent):
    name: str = field(init=False, default='muc-user-joined')
    timestamp: datetime.datetime
    is_self: bool
    nick: str
    status_codes: set[StatusCode] | None


@dataclass
class MUCUserLeft(ApplicationEvent):
    name: str = field(init=False, default='muc-user-left')
    timestamp: datetime.datetime
    is_self: bool
    nick: str
    status_codes: set[StatusCode] | None
    reason: str | None
    actor: str | None


@dataclass
class MUCUserRoleChanged(ApplicationEvent):
    name: str = field(init=False, default='muc-user-role-changed')
    timestamp: datetime.datetime
    is_self: bool
    nick: str
    role: Role
    reason: str | None
    actor: str | None


@dataclass
class MUCUserAffiliationChanged(ApplicationEvent):
    name: str = field(init=False, default='muc-user-affiliation-changed')
    timestamp: datetime.datetime
    is_self: bool
    nick: str
    affiliation: Affiliation
    reason: str | None
    actor: str | None


@dataclass
class MUCAffiliationChanged(ApplicationEvent):
    name: str = field(init=False, default='room-affiliation-changed')
    timestamp: datetime.datetime
    nick: str
    affiliation: Affiliation
    jid: JID


@dataclass
class MUCUserStatusShowChanged(ApplicationEvent):
    name: str = field(init=False, default='muc-user-status-show-changed')
    timestamp: datetime.datetime
    is_self: bool
    nick: str
    status: str
    show_value: str


@dataclass
class EncryptionInfo(ApplicationEvent):
    name: str = field(init=False, default='encryption-check')
    account: str
    jid: JID
    type: EncryptionInfoMsg
    message: str


@dataclass
class DBMigration(ApplicationEvent):
    name: str = field(init=False, default='db-migration')


@dataclass
class DBMigrationStart(ApplicationEvent):
    name: str = field(init=False, default='db-migration-start')
    version: int


@dataclass
class DBMigrationProgress(ApplicationEvent):
    name: str = field(init=False, default='db-migration-progress')
    count: int
    progress: int

    @property
    def value(self) -> str:
        value = 100
        if self.count != 0:
            value = self.progress / self.count * 100
        return f'{value:05.2f}'


@dataclass
class DBMigrationFinished(ApplicationEvent):
    name: str = field(init=False, default='db-migration-finished')


@dataclass
class DBMigrationError(ApplicationEvent):
    name: str = field(init=False, default='db-migration-error')
    exception: Exception


@dataclass
class ContactBlocked(ApplicationEvent):
    name: str = field(init=False, default='contact-blocked')
    account: str
    jid: JID
